from collections import OrderedDict
from typing import Optional, List, Tuple, Dict, Set, Union

from pikachu.drawing.drawing import *
from pikachu.math_functions import *
from pikachu.chem.atom import Atom
from pikachu.chem.bond import Bond

from raichu.central_chain_detection.find_central_chain import find_central_chain, find_central_chain_not_attached, \
    find_central_chain_ripp, reorder_central_chain
from raichu.central_chain_detection.label_central_chain import label_ripp_central_chain


def get_ring(atom, structure):
    cycles = structure.cycles.all_cycles

    for i, cycle in enumerate(cycles):
        if atom in cycle:
            return cycle

    return None


class RaichuDrawer(Drawer):
    def __init__(self, structure, options=None, save_png=None, dont_show=False,
                 coords_only=True, dpi=100, save_svg=None, draw_Cs_in_pink=False, add_url=False, horizontal=False,
                 ripp=False, make_linear=True):
        self.dont_show = dont_show
        self.dpi = dpi
        self.draw_Cs_in_pink = draw_Cs_in_pink
        self.add_url = add_url
        self.horizontal = horizontal
        self.ripp = ripp
        self.make_linear = make_linear
        self.line_width = 2

        if options is None:
            self.options = Options()
        else:
            self.options = options
        if save_png is None:
            self.save_png = None
        else:
            # Check if filename is valid
            assert save_png.endswith('.png')
            self.save_png = save_png
        if save_svg is None:
            self.save_svg = None
        else:
            # Check if filename is valid
            assert save_svg.endswith('.svg')
            self.save_svg = save_svg
        super().__init__(structure, options=None, coords_only=coords_only)

    def center_on_carrier_domain(self):
        pass

    def find_clashing_atoms(self) -> List[Tuple[Atom, Atom]]:
        clashing_atoms = []
        for i, atom_1 in enumerate(self.drawn_atoms):
            for j in range(i + 1, len(self.drawn_atoms)):
                atom_2 = self.drawn_atoms[j]
                if not self.structure.bond_exists(atom_1, atom_2):
                    distance = Vector.subtract_vectors(atom_1.draw.position, atom_2.draw.position).get_squared_length()
                    if atom_1.has_neighbour('H') and atom_2.has_neighbour('H') and atom_1.type != 'C' and \
                            atom_2.type != 'C':
                        max_distance = self.options.bond_length_squared * 1.5
                    # elif atom_1.has_neighbour('H') or atom_2.has_neighbour('H'):
                    #     max_distance = self.options.bond_length_squared / 1.5
                    else:
                        max_distance = self.options.bond_length_squared / 2
                    if distance <= max_distance:
                        clashing_atoms.append((atom_1, atom_2))

        return clashing_atoms
    
    def find_out_of_bound_atoms(self, minimum_y) -> List[Atom]:
        clashing_atoms = []
        for i, atom in enumerate(self.drawn_atoms):
            if atom.draw.position.y < minimum_y and atom.type != 'I':
                clashing_atoms.append(atom)
        return clashing_atoms

    # TODO: add an option in pikachu Options object to set the min bond distance between which to consider fine-tuning
    def finetune_overlap_resolution(self, masked_bonds=None, highest_atom=None):

        if not masked_bonds:
            masked_bonds = set()

        else:
            masked_bonds = set(masked_bonds)

        # if self.total_overlap_score > self.options.overlap_sensitivity:
        clashing_atoms = self.find_clashing_atoms()

        best_bonds = []
        for atom_1, atom_2 in clashing_atoms:

            shortest_path = self.find_shortest_path(atom_1, atom_2)
            rotatable_bonds = []
            distances = []

            for i, bond in enumerate(shortest_path):
                distance_1 = i
                distance_2 = len(shortest_path) - i

                average_distance = len(shortest_path) / 2

                distance_metric = abs(
                    average_distance - distance_1) + abs(average_distance - distance_2)

                if self.bond_is_rotatable(bond) and bond not in masked_bonds:
                    rotatable_bonds.append(bond)
                    distances.append(distance_metric)

            best_bond = None
            optimal_distance = float('inf')
            for i, distance in enumerate(distances):
                if distance < optimal_distance:
                    # if highest_atom:
                    #     if atom_1.draw.position.y < highest_atom.draw.position.y and \
                    #             atom_2.draw.position.y < highest_atom.draw.position.y:
                    #         best_bond = rotatable_bonds[i]
                    #         optimal_distance = distance
                    # else:
                        best_bond = rotatable_bonds[i]
                        optimal_distance = distance

            if best_bond:
                best_bonds.append(best_bond)

        best_bonds = list(set(best_bonds))

        for best_bond in best_bonds:
            if self.total_overlap_score > self.options.overlap_sensitivity:

                subtree_size_1 = self.get_subgraph_size(
                    best_bond.atom_1, {best_bond.atom_2})
                subtree_size_2 = self.get_subgraph_size(
                    best_bond.atom_2, {best_bond.atom_1})

                if subtree_size_1 < subtree_size_2:
                    rotating_atom = best_bond.atom_1
                    parent_atom = best_bond.atom_2
                else:
                    rotating_atom = best_bond.atom_2
                    parent_atom = best_bond.atom_1

                overlap_score, _, _ = self.get_overlap_score()

                self.rotate_subtree(rotating_atom, parent_atom, math.radians(
                    15), parent_atom.draw.position)

                new_overlap_score, _, _ = self.get_overlap_score()

                scores = [new_overlap_score]

                for i in range(12):
                    self.rotate_subtree(rotating_atom, parent_atom, math.radians(
                        30), parent_atom.draw.position)
                    new_overlap_score, _, _ = self.get_overlap_score()
                    scores.append(new_overlap_score)

                assert len(scores) == 13
                scores = scores[:12]

                best_i = 0
                best_score = scores[0]

                for i, score in enumerate(scores):
                    if score < best_score:
                        best_score = score
                        best_i = i

                self.total_overlap_score = best_score

                self.rotate_subtree(rotating_atom, parent_atom, math.radians(
                    30 * best_i + 1), parent_atom.draw.position)
    
    def finetune_overlap_bubbles(self, masked_bonds=None, highest_atom=None):

        if not masked_bonds:
            masked_bonds = set()

        else:
            masked_bonds = set(masked_bonds)

        domain = None
        
        # Find minimum y

        for atom in self.structure.graph:
            if atom.type == 'I':
                if atom.annotations.domain_type in ["Leader", "Follower", "ACP", "PCP"]:
                    domain = atom
        # Add also padding for hydrogens drawn as a atom label but not as an explicit hydrogen (e.g. HN)
        assert domain is not None

        minimum_y = domain.draw.position.y + 5
        
        clashing_atoms = self.find_out_of_bound_atoms(minimum_y)

        best_bonds = []
        for atom_1 in clashing_atoms:

            shortest_path = self.find_shortest_path(atom_1, domain)
            rotatable_bonds = []
            distances = []

            for i, bond in enumerate(shortest_path):
                distance_1 = i
                distance_2 = len(shortest_path) - i

                average_distance = len(shortest_path) / 2

                distance_metric = abs(
                    average_distance - distance_1) + abs(average_distance - distance_2)

                if self.bond_is_rotatable(bond) and bond not in masked_bonds:
                    rotatable_bonds.append(bond)
                    distances.append(distance_metric)

            best_bond = None
            optimal_distance = float('inf')
            for i, distance in enumerate(distances):
                if distance < optimal_distance:
                    # if highest_atom:
                    #     if atom_1.draw.position.y < highest_atom.draw.position.y and \
                    #             atom_2.draw.position.y < highest_atom.draw.position.y:
                    #         best_bond = rotatable_bonds[i]
                    #         optimal_distance = distance
                    # else:
                    best_bond = rotatable_bonds[i]
                    optimal_distance = distance

            if best_bond:
                best_bonds.append(best_bond)

        best_bonds = list(set(best_bonds))

        for best_bond in best_bonds:
            if len(self.find_out_of_bound_atoms(minimum_y))>0:

                path1 = self.find_shortest_path(best_bond.atom_1, domain)
                path2 = self.find_shortest_path(best_bond.atom_2, domain)
                
                if len(path2) < len(path1):
                    rotating_atom = best_bond.atom_1
                    parent_atom = best_bond.atom_2
                else:
                    rotating_atom = best_bond.atom_2
                    parent_atom = best_bond.atom_1

                new_overlap_score, _, _ = self.get_overlap_score()

                scores = [new_overlap_score]
                overlappings = [len(self.find_out_of_bound_atoms(minimum_y))]

                self.rotate_subtree(rotating_atom, parent_atom, math.radians(
                    15), parent_atom.draw.position)

                for i in range(12):
                    self.rotate_subtree(rotating_atom, parent_atom, math.radians(
                        30), parent_atom.draw.position)
                    overlappings.append(len(self.find_out_of_bound_atoms(minimum_y)))
                    new_overlap_score, _, _ = self.get_overlap_score()
                    scores.append(new_overlap_score)

                assert len(scores) == 13
                scores = scores[:12]

                assert len(overlappings) == 13
                overlappings = overlappings[:12]

                best_i = 0
                best_overlapping = overlappings[0]
                best_score = self.total_overlap_score
                for i, overlapping in enumerate(overlappings):
                    if overlapping < best_overlapping and scores[i] < self.options.overlap_sensitivity + self.total_overlap_score:
                        best_overlapping = overlapping
                        best_i = i
                        best_score = scores[i]
                self.rotate_subtree(rotating_atom, parent_atom, math.radians(
                    30 * best_i + 1), parent_atom.draw.position)
                self.total_overlap_score = best_score
                

    def draw(self, coords_only: bool = False) -> None:

        if not self.options.draw_hydrogens:
            self.hide_hydrogens()

        self.get_atom_nr_to_atom()
        self.define_rings()

        if not self.multiple:
            self._process_structure()
            if self.make_linear:
                self.linearise()
            self.set_chiral_bonds()
            if not coords_only:
                self.draw_structure()
        else:
            self.restore_ring_information()

    # TODO: replace this entirely with PIKAChU's svg drawer
    def draw_structure(self):
        # Find the plotting dimensions of the molecule such that the canvas can be scaled to fit the molecule

        min_x = 100000000
        max_x = -100000000
        min_y = 100000000
        max_y = -100000000

        for atom in self.structure.graph:
            if atom.draw.positioned:
                if atom.draw.position.x < min_x:
                    min_x = atom.draw.position.x
                if atom.draw.position.y < min_y:
                    min_y = atom.draw.position.y
                if atom.draw.position.x > max_x:
                    max_x = atom.draw.position.x
                if atom.draw.position.y > max_y:
                    max_y = atom.draw.position.y

        height = max_y - min_y
        width = max_x - min_x

        fig, ax = plt.subplots(figsize=((width + 2 * self.options.padding) / 50.0,
                                        (height + 2 * self.options.padding) / 50.0), dpi=100)

        ax.set_aspect('equal', adjustable='box')
        ax.axis('off')

        ax.set_xlim([min_x - self.options.padding, max_x + self.options.padding])
        ax.set_ylim([min_y - self.options.padding, max_y + self.options.padding])
        plt.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)

        params = {'mathtext.default': 'regular'}
        plt.rcParams.update(params)

        ring_centers_x = []
        ring_centers_y = []

        for ring in self.rings:
            self.set_ring_center(ring)

            ring_centers_x.append(ring.center.x)
            ring_centers_y.append(ring.center.y)

        for bond_nr, bond in self.structure.bonds.items():
            if bond.atom_1.draw.positioned and bond.atom_2.draw.positioned:
                line = Line(bond.atom_1.draw.position, bond.atom_2.draw.position, bond.atom_1, bond.atom_2)
                midpoint = line.get_midpoint()

                if bond.type == 'single':
                    if bond in self.chiral_bonds:
                        orientation, chiral_center = self.chiral_bond_to_orientation[bond]
                        self._plot_chiral_bond(orientation, chiral_center, line, ax, midpoint)
                    else:
                        self._plot_halflines(line, ax, midpoint)
                elif bond.type == 'double':
                    if not self._is_terminal(bond.atom_1) and not self._is_terminal(bond.atom_2):
                        self._plot_halflines(line, ax, midpoint)

                        common_ring_numbers = self._get_common_rings(bond.atom_1, bond.atom_2)

                        if common_ring_numbers:
                            common_rings = []
                            for ring_nr in common_ring_numbers:
                                common_rings.append(self.get_ring(ring_nr))

                            common_rings.sort(key=lambda x: len(x.members))
                            common_ring = common_rings[0]
                            ring_centre = common_ring.center
                            second_line = line.double_line_towards_center(ring_centre, self.options.bond_spacing,
                                                                          self.options.double_bond_length)
                            second_line_midpoint = second_line.get_midpoint()
                            self._plot_halflines_double(second_line, ax, second_line_midpoint)

                        else:
                            bond_neighbours = bond.atom_1.drawn_neighbours + bond.atom_2.drawn_neighbours
                            if bond_neighbours:
                                vectors = [atom.draw.position for atom in bond_neighbours]
                                gravitational_point = Vector.get_average(vectors)
                                second_line = line.double_line_towards_center(gravitational_point,
                                                                              self.options.bond_spacing,
                                                                              self.options.double_bond_length)
                                second_line_midpoint = second_line.get_midpoint()
                                self._plot_halflines_double(second_line, ax, second_line_midpoint)
                            else:
                                print("Shouldn't happen!")
                    else:
                        if self._is_terminal(bond.atom_1) and self._is_terminal(bond.atom_2):
                            dummy_1 = Vector(bond.atom_1.draw.position.x + 1, bond.atom_1.draw.position.y + 1)
                            dummy_2 = Vector(bond.atom_1.draw.position.x - 1, bond.atom_1.draw.position.y - 1)
                            double_bond_line_1 = line.double_line_towards_center(dummy_1,
                                                                                 self.options.bond_spacing / 2.0,
                                                                                 self.options.double_bond_length)
                            double_bond_line_1_midpoint = double_bond_line_1.get_midpoint()
                            double_bond_line_2 = line.double_line_towards_center(dummy_2,
                                                                                 self.options.bond_spacing / 2.0,
                                                                                 self.options.double_bond_length)
                            double_bond_line_2_midpoint = double_bond_line_2.get_midpoint()

                            self._plot_halflines_double(double_bond_line_1, ax, double_bond_line_1_midpoint)
                            self._plot_halflines_double(double_bond_line_2, ax, double_bond_line_2_midpoint)

                        else:

                            if self._is_terminal(bond.atom_1):
                                terminal_atom = bond.atom_1
                                branched_atom = bond.atom_2
                            else:
                                terminal_atom = bond.atom_2
                                branched_atom = bond.atom_1

                            if len(branched_atom.drawn_neighbours) >= 3:
                                closest_two = self.get_sorted_distances_from_list(terminal_atom,
                                                                                  branched_atom.drawn_neighbours)
                                closest_atom_1 = closest_two[0][1]
                                closest_atom_2 = closest_two[1][1]

                                line = Line(terminal_atom.draw.position, branched_atom.draw.position, terminal_atom,
                                            branched_atom)

                                double_bond_line_1, double_bond_line_2 = line.get_perpendicular_lines(
                                    self.options.bond_spacing / 2.0)
                                terminal_atom_pos_1 = double_bond_line_1.get_atom_coords(terminal_atom)
                                terminal_atom_pos_2 = double_bond_line_2.get_atom_coords(terminal_atom)

                                closest_atom_to_pos_1 = terminal_atom_pos_1.get_closest_atom(closest_atom_1,
                                                                                             closest_atom_2)
                                closest_atom_to_pos_2 = terminal_atom_pos_2.get_closest_atom(closest_atom_1,
                                                                                             closest_atom_2)

                                bond_1_line = Line(branched_atom.draw.position, closest_atom_to_pos_1.draw.position,
                                                   branched_atom, closest_atom_to_pos_1)
                                bond_2_line = Line(branched_atom.draw.position, closest_atom_to_pos_2.draw.position,
                                                   branched_atom, closest_atom_to_pos_2)

                                double_bond_line_1_midpoint = double_bond_line_1.get_midpoint()
                                double_bond_line_2_midpoint = double_bond_line_2.get_midpoint()

                                intersection_1 = double_bond_line_1.find_intersection(bond_1_line)
                                intersection_2 = double_bond_line_2.find_intersection(bond_2_line)

                                if terminal_atom.draw.position.x > branched_atom.draw.position.x:
                                    # check for parallel lines
                                    if intersection_1 and intersection_1.x < 100000 and intersection_1.y < 100000:
                                        double_bond_line_1.point_1 = intersection_1
                                    if intersection_2 and intersection_2.x < 100000 and intersection_2.y < 100000:
                                        double_bond_line_2.point_1 = intersection_2

                                else:
                                    # check for parallel lines
                                    if intersection_1 and intersection_1.x < 100000 and intersection_1.y < 100000:
                                        double_bond_line_1.point_2 = intersection_1
                                    if intersection_2 and intersection_2.x < 100000 and intersection_2.y < 100000:
                                        double_bond_line_2.point_2 = intersection_2

                                self._plot_halflines(double_bond_line_1, ax, double_bond_line_1_midpoint)
                                self._plot_halflines(double_bond_line_2, ax, double_bond_line_2_midpoint)

                            else:
                                self._plot_halflines(line, ax, midpoint)

                                bond_neighbours = bond.atom_1.drawn_neighbours + bond.atom_2.drawn_neighbours
                                if bond_neighbours:
                                    vectors = [atom.draw.position for atom in bond_neighbours]
                                    gravitational_point = Vector.get_average(vectors)
                                    second_line = line.get_parallel_line(gravitational_point,
                                                                         self.options.bond_spacing)
                                    second_line_midpoint = second_line.get_midpoint()
                                    self._plot_halflines(second_line, ax, second_line_midpoint)
                                else:
                                    print("Shouldn't happen!")

                elif bond.type == 'triple':
                    self._plot_halflines(line, ax, midpoint)
                    line_1, line_2 = line.get_parallel_lines(self.options.bond_spacing)
                    line_1_midpoint = line_1.get_midpoint()
                    line_2_midpoint = line_2.get_midpoint()
                    self._plot_halflines(line_1, ax, line_1_midpoint)
                    self._plot_halflines(line_2, ax, line_2_midpoint)

        # If the starter unit contains an unknown moiety, number this R1
        for atom in self.structure.graph:
            if atom.type == '*' and not atom.annotations.unknown_index:
                atom.annotations.unknown_index = 1
        for atom in self.structure.graph:
            if ((atom.type != 'C' or self.draw_Cs_in_pink) and getattr(atom.annotations,"domain_type", None) not in ["Follower", "Leader"]) and atom.draw.positioned:
                text_h = ''
                text_h_pos = None

                if (atom.type != 'C' or self.draw_Cs_in_pink or atom.draw.draw_explicit) :
                    text = atom.type
                else:
                    text = ''

                if hasattr(atom.annotations, "domain_type") and atom.annotations.domain_type:
                    if atom.annotations.domain_type in ["Follower", "Leader"]:
                        text = ""
                    else:
                        text = atom.annotations.domain_type

                horizontal_alignment = 'center'

                orientation = self._get_hydrogen_text_orientation(atom)
                if orientation == 'H_above_atom':
                    text_h_pos = Vector(
                        atom.draw.position.x, atom.draw.position.y + 6)
                if orientation == 'H_below_atom':
                    text_h_pos = Vector(
                        atom.draw.position.x, atom.draw.position.y - 6)

                atom_draw_position = Vector(
                    atom.draw.position.x, atom.draw.position.y)

                if atom.type == '*':
                    neighbouring_c = atom.get_neighbour('C')
                    assert neighbouring_c
                    # In order to not let the number of the sidechain overlap
                    # with the bond, move Rx symbol along hor. axis depending on
                    # if group is added to the left or right of the chain
                    delta_x_r = self.get_delta_x_sidechain(
                        atom, neighbouring_c)
                    atom_draw_position.x += delta_x_r
                    text = fr'$R_{atom.annotations.unknown_index}$'

                if not atom.charge and (atom.type != 'C' or self.draw_Cs_in_pink or atom.draw.draw_explicit) and getattr(atom.annotations, "domain_type", None) not in ["Follower", "Leader"]:

                    if atom.draw.has_hydrogen:
                        hydrogen_count = 0
                        for neighbour in atom.neighbours:
                            if neighbour.type == 'H' and not neighbour.draw.is_drawn:
                                hydrogen_count += 1

                        if hydrogen_count and not (atom.type == 'C' and self.draw_Cs_in_pink or getattr(atom.annotations, "domain_type", None) in ["Follower", "Leader"]):

                            if hydrogen_count > 1:
                                if orientation == 'H_before_atom':
                                    text = r'$H_{hydrogens}{atom_type}$'.format(hydrogens=hydrogen_count,
                                                                                atom_type=atom.type)
                                    horizontal_alignment = 'right'
                                    atom_draw_position.x += 3
                                elif orientation == 'H_below_atom' or orientation == 'H_above_atom':
                                    text = atom.type
                                    text_h = r'$H_{hydrogens}$'.format(
                                        hydrogens=hydrogen_count)

                                else:
                                    text = r'${atom_type}H_{hydrogens}$'.format(hydrogens=hydrogen_count,
                                                                                atom_type=atom.type)
                                    horizontal_alignment = 'left'
                                    atom_draw_position.x -= 3
                            elif hydrogen_count == 1:
                                if orientation == 'H_before_atom':
                                    text = f'H{atom.type}'
                                    horizontal_alignment = 'right'
                                    atom_draw_position.x += 3
                                elif orientation == 'H_below_atom' or orientation == 'H_above_atom':
                                    text = atom.type
                                    text_h = 'H'
                                else:
                                    text = f'{atom.type}H'
                                    horizontal_alignment = 'left'
                                    atom_draw_position.x -= 3

                elif atom.charge:
                    if atom.charge > 0:
                        charge_symbol = '+'
                    else:
                        charge_symbol = '-'

                    hydrogen_count = 0
                    for neighbour in atom.neighbours:
                        if neighbour.type == 'H' and not neighbour.draw.is_drawn:
                            hydrogen_count += 1

                    if not hydrogen_count:

                        if abs(atom.charge) > 1:

                            text = r'${atom_type}^{charge}{charge_symbol}$'.format(charge=atom.charge,
                                                                                   atom_type=atom.type,
                                                                                   charge_symbol=charge_symbol)
                        elif abs(atom.charge) == 1:
                            text = r'${atom_type}^{charge_symbol}$'.format(atom_type=atom.type,
                                                                           charge_symbol=charge_symbol)

                        horizontal_alignment = 'left'
                        atom_draw_position.x -= 3
                    else:

                        if hydrogen_count > 1:
                            if orientation == 'H_before_atom':
                                if abs(atom.charge) > 1:
                                    text = r'$H_{hydrogens}{atom_type}^{charge}{charge_symbol}$'.format(
                                        hydrogens=hydrogen_count,
                                        atom_type=atom.type,
                                        charge=atom.charge,
                                        charge_symbol=charge_symbol)
                                elif abs(atom.charge) == 1:
                                    text = r'$H_{hydrogens}{atom_type}^{charge_symbol}$'.format(
                                        hydrogens=hydrogen_count,
                                        atom_type=atom.type,
                                        charge_symbol=charge_symbol)

                                horizontal_alignment = 'right'
                                atom_draw_position.x += 3
                            elif orientation == 'H_above_atom' or orientation == 'H_below_atom':
                                text_h = r'$H_{hydrogens}$'.format(
                                    hydrogens=hydrogen_count)
                                if abs(atom.charge) > 1:
                                    text = r'${atom_type}^{charge}{charge_symbol}$'.format(atom_type=atom.type,
                                                                                           charge=atom.charge,
                                                                                           charge_symbol=charge_symbol)
                                elif abs(atom.charge) == 1:
                                    text = r'${atom_type}^{charge_symbol}$'.format(atom_type=atom.type,
                                                                                   charge_symbol=charge_symbol)
                            else:
                                if abs(atom.charge) > 1:
                                    text = r'${atom_type}H_{hydrogens}^{charge}{charge_symbol}$'.format(
                                        hydrogens=hydrogen_count,
                                        atom_type=atom.type,
                                        charge=atom.charge,
                                        charge_symbol=charge_symbol)
                                elif abs(atom.charge) == 1:
                                    text = r'${atom_type}H_{hydrogens}^{charge_symbol}$'.format(
                                        hydrogens=hydrogen_count,
                                        atom_type=atom.type,
                                        charge_symbol=charge_symbol)

                                horizontal_alignment = 'left'
                                atom_draw_position.x -= 3
                        elif hydrogen_count == 1:
                            if orientation == 'H_before_atom':
                                if abs(atom.charge) > 1:

                                    text = r'$H{atom_type}^{charge}{charge_symbol}$'.format(atom_type=atom.type,
                                                                                            charge=atom.charge,
                                                                                            charge_symbol=charge_symbol)
                                elif abs(atom.charge) == 1:
                                    text = r'$H{atom_type}^{charge_symbol}$'.format(atom_type=atom.type,
                                                                                    charge_symbol=charge_symbol)
                                horizontal_alignment = 'right'
                                atom_draw_position.x += 3
                            elif orientation == 'H_above_atom' or orientation == 'H_below_atom':
                                text_h = 'H'
                                if abs(atom.charge) > 1:

                                    text = r'${atom_type}^{charge}{charge_symbol}$'.format(atom_type=atom.type,
                                                                                           charge=atom.charge,
                                                                                           charge_symbol=charge_symbol)
                                elif abs(atom.charge) == 1:
                                    text = r'${atom_type}^{charge_symbol}$'.format(atom_type=atom.type,
                                                                                   charge_symbol=charge_symbol)

                            else:
                                if abs(atom.charge) > 1:
                                    text = r'${atom_type}H^{charge}{charge_symbol}$'.format(atom_type=atom.type,
                                                                                            charge=atom.charge,
                                                                                            charge_symbol=charge_symbol)

                                elif abs(atom.charge) == 1:
                                    text = r'${atom_type}H^{charge_symbol}$'.format(atom_type=atom.type,
                                                                                    charge_symbol=charge_symbol)
                                horizontal_alignment = 'left'
                                atom_draw_position.x -= 3

                atom_color = atom.draw.colour
                if self.draw_Cs_in_pink:
                    if atom.type == 'C':
                        atom_color = "magenta"
                if self.add_url:
                    if text:
                        plt.text(atom_draw_position.x, atom_draw_position.y,
                                 text, url=str(atom),
                                 horizontalalignment=horizontal_alignment,
                                 verticalalignment='center',
                                 color=atom_color)
                    if text_h:
                        plt.text(text_h_pos.x, text_h_pos.y,
                                 text_h, url=str(atom),
                                 horizontalalignment='center',
                                 verticalalignment='center',
                                 color=atom_color)
                else:
                    if text:
                        plt.text(atom_draw_position.x, atom_draw_position.y,
                                 text,
                                 horizontalalignment=horizontal_alignment,
                                 verticalalignment='center',
                                 color=atom_color)
                    if text_h:
                        plt.text(text_h_pos.x, text_h_pos.y,
                                 text_h,
                                 horizontalalignment='center',
                                 verticalalignment='center',
                                 color=atom_color)

        # If a png filename is included in the initialization of the
        # Raichu_drawer object, don't show the structure, but do save it as a
        # png image to the provided filename
        if self.dont_show:
            return self

        elif self.save_png is None and not self.dont_show and self.save_svg is None:
            plt.show()

        else:
            if self.save_png:
                plt.savefig(self.save_png)
            elif self.save_svg:
                plt.savefig(self.save_svg)
            plt.clf()
            plt.close()

    def colour_modules(self):
        pass

    def place_top_atom(self, backbone_atoms, attachment_point):


        # Fix position of the S and C atom and the S-C angle
        top_atom = backbone_atoms[0]
        first_carbon = backbone_atoms[1]

        top_atom.draw.position.x = first_carbon.draw.position.x
        top_atom.draw.position.y = first_carbon.draw.position.y + 15

        angle = get_angle(first_carbon.draw.position,
                          top_atom.draw.position)

        angle_degrees = round(math.degrees(angle), 3)

        correct_angle_deg = -120
        delta_angle_deg = correct_angle_deg - angle_degrees
        delta_angle_rad = math.radians(delta_angle_deg)

        self.rotate_subtree(top_atom, first_carbon, delta_angle_rad,
                            first_carbon.draw.position)

        # Fix position domain straight above sulphur atom
        attachment_point.draw.position.x = top_atom.draw.position.x
        attachment_point.draw.position.y = top_atom.draw.position.y + 15


    @staticmethod
    def get_opposite_placement(placement):
        if placement == 'right':
            return 'left'
        elif placement == 'left':
            return 'right'
        else:
            raise ValueError("Only placements accepted are 'left' and 'right'.")

    def get_placements(self, backbone, atom_to_ring, stop_linearising):
        backbone_to_placement = OrderedDict()
        for i, atom in enumerate(backbone):
            if atom == stop_linearising:
                break

            if i == 0:
                backbone_to_placement[atom] = 'right'
            elif i == 1:
                backbone_to_placement[atom] = 'left'
            else:
                if atom not in backbone_to_placement:
                    previous_placement = backbone_to_placement.get(backbone[i - 1])
                    ring = atom_to_ring.get(atom)
                    if ring and len(ring) == 2:
                        if ring[0] in backbone_to_placement:
                            backbone_to_placement[ring[1]] = backbone_to_placement.get(ring[0])
                        else:
                            backbone_to_placement[atom] = self.get_opposite_placement(previous_placement)
                    elif hasattr(atom, "domain_type") and atom.annotations.domain_type == "Leader":
                        backbone_to_placement[atom] = previous_placement
                    else:
                        backbone_to_placement[atom] = self.get_opposite_placement(previous_placement)

        return backbone_to_placement

    def fix_rings(self, rings, backbone_atoms, backbone_to_placement):
        for ring in rings:
            if len(ring) == 1:
                atom_1 = ring[0]
                if atom_1 == backbone_atoms[-1]:
                    continue
                drawing_ring = self.get_ring(atom_1.draw.rings[0])
                center = Vector(0, 0)
                for atom in drawing_ring.members:
                    center.add(atom.draw.position)

                center.divide(len(drawing_ring.members))
                angle = get_angle(atom_1.draw.position,
                                  center)
                
                if backbone_to_placement[atom_1] == 'right':
                    desired_angle = 180
                else:
                    desired_angle = 0.0
                required_rotation_deg = desired_angle - angle
                required_rotation_rad = math.radians(required_rotation_deg)

                masked = {atom_1}
                first_atom_cycle = None
                for next_atom in atom_1.neighbours:
                    if next_atom.type != 'H' and next_atom not in backbone_atoms and next_atom in drawing_ring.members:
                        first_atom_cycle = next_atom

                assert first_atom_cycle

                # for atom in drawing_ring.members:
                for atom in self.traverse_substructure(first_atom_cycle, masked):
                    if atom != atom_1:
                        atom.draw.position.rotate_around_vector(required_rotation_rad, atom_1.draw.position)

            elif len(ring) == 2:
                atom_1, atom_2 = ring
                masked = {atom_1, atom_2}

                first_atom_cycle = None
                for next_atom in atom_1.neighbours:
                    if next_atom.type != 'H' and next_atom not in backbone_atoms:
                        first_atom_cycle = next_atom

                assert first_atom_cycle

                if backbone_to_placement[atom_1] == backbone_to_placement[atom_2] == 'right':
                    if atom_1.draw.position.x > first_atom_cycle.draw.position.x:
                        for atom in self.traverse_substructure(first_atom_cycle, masked):
                            delta_x = 2 * (atom_1.draw.position.x - atom.draw.position.x)
                            atom.draw.position.x += delta_x

                elif backbone_to_placement[atom_1] == backbone_to_placement[atom_2] == 'left':

                    if atom_1.draw.position.x < first_atom_cycle.draw.position.x:
                        for atom in self.traverse_substructure(first_atom_cycle, masked):
                            delta_x = 2 * (atom.draw.position.x - atom_1.draw.position.x)
                            atom.draw.position.x -= delta_x

    def get_overlap_score(self) -> Tuple[float, List[Tuple[float, Atom]], Dict[Atom, float]]:
        total = 0.0

        overlap_scores = {}
        for atom in self.drawn_atoms:
            overlap_scores[atom] = 0.0

        for i, atom_1 in enumerate(self.drawn_atoms):
            for j in range(i + 1, len(self.drawn_atoms)):
                atom_2 = self.drawn_atoms[j]
                distance = Vector.subtract_vectors(atom_1.draw.position, atom_2.draw.position).get_squared_length()
                if atom_1 not in atom_2.neighbours:
                    if distance < self.options.bond_length_squared + 1:

                        if atom_1.has_neighbour('H') and atom_2.has_neighbour('H') and atom_1.type != 'C' \
                                and atom_2.type != 'C':

                            weight = self.options.overlap_sensitivity + abs((self.options.bond_length - math.sqrt(distance))) / self.options.bond_length

                        else:
                            weight = abs((self.options.bond_length - math.sqrt(distance))) / self.options.bond_length
                        total += weight
                        overlap_scores[atom_1] += weight
                        overlap_scores[atom_2] += weight

        sorted_overlaps = []

        for atom in self.drawn_atoms:
            sorted_overlaps.append((overlap_scores[atom], atom))

        sorted_overlaps.sort(key=lambda x: x[0], reverse=True)

        return total, sorted_overlaps, overlap_scores

    def linearise(self):

        attachment_point = None
        domains = []

        for atom in self.structure.graph:
            if atom.type == 'I':
                if atom.annotations.domain_type in ["Leader", "Follower", "ACP", "PCP"]:
                    domains.append(atom)
                    # Ensures the follower peptide is set as attachment point if both leader and follower are present
                    if not attachment_point:
                        attachment_point = atom
                    elif attachment_point.annotations.domain_type == "Leader":
                        attachment_point = atom

        if attachment_point and attachment_point.annotations.domain_type == "Leader":
            horizontal_rotation = 'anticlockwise'
        else:
            horizontal_rotation = 'clockwise'

        if self.ripp:
            label_ripp_central_chain(self.structure)

        self.structure.refresh_structure()

        if attachment_point:
            if self.ripp:
                backbone_atoms = find_central_chain_ripp(self.structure)
            else:
                backbone_atoms = find_central_chain(self.structure)
        else:
            backbone_atoms = find_central_chain_not_attached(self.structure)

        if attachment_point:

            self.place_top_atom(backbone_atoms, attachment_point)

        self.structure.refresh_structure()
        backbone, full_rings, rings, atom_to_ring, stop_linearising = reorder_central_chain(backbone_atoms, self)
        if len(domains) == 2:
            for domain in domains:
                if domain.annotations.domain_type == 'Leader':
                    backbone.append(domain)

        backbone_to_placement = self.get_placements(backbone, atom_to_ring, stop_linearising)

        i = 0

        while i < len(backbone) - 1:

            atom_1 = backbone[i]
            atom_2 = backbone[i + 1]

            if atom_1 not in backbone_to_placement or atom_2 not in backbone_to_placement:
                break

            current_angle = get_angle(atom_1.draw.position, atom_2.draw.position)

            if backbone_to_placement[atom_1] == 'right' and backbone_to_placement[atom_2] == 'left':
                desired_angle = 60.0
            elif backbone_to_placement[atom_1] == 'left' and backbone_to_placement[atom_2] == 'right':
                desired_angle = 120.0
            elif backbone_to_placement[atom_1] == backbone_to_placement[atom_2]:
                desired_angle = 90.0
            else:
                raise ValueError("Only supported orientations are 'left' and 'right'.")

            required_rotation_deg = desired_angle - current_angle
            required_rotation_rad = math.radians(required_rotation_deg)

            if not (atom_2 in atom_to_ring and atom_1 in atom_to_ring and atom_to_ring[atom_2] == atom_to_ring[atom_1]):

                self.rotate_subtree(atom_2, atom_1, required_rotation_rad,
                                    atom_1.draw.position)
            else:
                if len(atom_to_ring[atom_2]) == 2:
                    self.rotate_subtree(atom_2, atom_1, required_rotation_rad,
                                        atom_1.draw.position)
                elif len(atom_to_ring[atom_2]) == 3:
                    if atom_2 == atom_to_ring[atom_2][1]:

                        self.rotate_subtree(atom_2, atom_1, required_rotation_rad,
                                            atom_1.draw.position)
                    elif atom_2 == atom_to_ring[atom_2][2]:

                        ring = atom_to_ring[atom_2]
                        ring_atoms = []
                        sidechain_atoms = []
                        for atom in atom_2.get_ring(self.structure):
                            if atom not in backbone:
                                ring_atoms.append(atom)

                        if abs(required_rotation_deg) > 90:
                            for atom in ring_atoms:
                                for neighbour in atom.neighbours:
                                    if neighbour.type != 'H':
                                        for sidechain_atom in self.traverse_substructure(neighbour,
                                                                                         set(ring_atoms + backbone)):
                                            if sidechain_atom not in sidechain_atoms and \
                                                    sidechain_atom not in ring_atoms and sidechain_atom not in backbone:
                                                sidechain_atoms.append(sidechain_atom)

                            for ring_atom in ring_atoms:
                                ring_atom.draw.position.mirror_about_line(ring[0].draw.position, ring[1].draw.position)
                            for sidechain_atom in sidechain_atoms:
                                sidechain_atom.draw.position.mirror_about_line(ring[0].draw.position, ring[1].draw.position)
                        self.rotate_subtree_ring(atom_2, set(ring_atoms + [atom_1]), required_rotation_rad,
                                                 atom_1.draw.position)

            i += 1

        atoms_in_rings = []
        for ring in full_rings:
            for atom in ring:
                atoms_in_rings.append(atom)
        atoms_in_rings = set(atoms_in_rings)

        self.fix_rings(rings, backbone, backbone_to_placement)

        self.position_sidechains(backbone, backbone_to_placement, atoms_in_rings)

        # self.resolve_primary_overlaps()
        self.total_overlap_score, sorted_overlap_scores, atom_to_scores = self.get_overlap_score()
        central_chain_bonds = set()

        for bond in self.structure.bonds.values():
            # if bond.atom_1.annotations.in_central_chain or \
            #         bond.atom_2.annotations.in_central_chain:
            if bond.atom_1 in backbone or bond.atom_2 in backbone:
                central_chain_bonds.add(bond)

        self.finetune_overlap_resolution(
            masked_bonds=central_chain_bonds, highest_atom=backbone[0])

        self.resolve_secondary_overlaps(sorted_overlap_scores)

        if self.horizontal:
            if horizontal_rotation == 'clockwise':
                self.rotate_structure(-1.5707)
            else:
                self.rotate_structure(1.5707)

    def resolve_overlaps(self, linear: bool = True, masked_bonds: Optional[Set["Bond"]] = None) -> None:
        """

        Parameters
        ----------
        linear: bool, if True, resolve overlaps for a linearised molecule; if False,
            resolve overlaps for a non-linearised molecule
        masked_bonds: set of bonds, bonds that are not allowed to be rotated

        """
        if not linear:

            self.resolve_primary_overlaps()
            self.total_overlap_score, sorted_overlap_scores, atom_to_scores = self.get_overlap_score()
            self.finetune_overlap_resolution()
            self.resolve_secondary_overlaps(sorted_overlap_scores)

        else:

            pass

    def find_high_atoms(self, atoms: List[Atom], max_y: float) -> List[Atom]:
        """
        Returns atoms that are drawn too high on the plot
        Parameters
        ----------
        atoms: list of [Atom, ->]
        max_y: float, maximum vertical y coordinate an atom is allowed to have

        Returns
        -------
        high_atoms: list of [Atom, ->], atoms with a y coordinate higher than max_y
        """
        high_atoms: List[Atom] = []
        for atom in atoms:
            if atom.draw.position.y > max_y:
                high_atoms.append(atom)
        return high_atoms

    def find_clashing_sidechain_atoms(self, sidechain_atoms):
        clashing_atoms: List[Tuple[Atom, Atom]] = []
        for i, atom_1 in enumerate(self.drawn_atoms):
            for j in range(i + 1, len(self.drawn_atoms)):
                atom_2 = self.drawn_atoms[j]
                if atom_1 in sidechain_atoms and not self.structure.bond_exists(atom_1, atom_2):
                    distance = Vector.subtract_vectors(atom_1.draw.position, atom_2.draw.position).get_squared_length()
                    if atom_1.has_neighbour('H') and atom_2.has_neighbour('H') and atom_1.type != 'C' and \
                            atom_2.type != 'C':
                        max_distance = self.options.bond_length_squared * 1.5
                    elif atom_1.has_neighbour('H') or atom_2.has_neighbour('H'):
                        max_distance = self.options.bond_length_squared
                    else:
                        max_distance = self.options.bond_length_squared * 0.8
                    if distance <= max_distance:
                        clashing_atoms.append((atom_1, atom_2))

        return clashing_atoms

    def resolve_sidechain_overlaps(self, backbone, backbone_to_placement, backbone_rings, central_chain_bonds):
        max_y = backbone[0].draw.position.y
        placed_atoms: List[Atom] = []
        backbone_to_sidechain: Dict[Atom, List[List[Atom]]] = {}
        backbone_to_rotatable_bonds: Dict[Atom, List[Bond]] = {}

        for backbone_atom in backbone:
            sidechains = self.get_sidechains(backbone, backbone_atom, backbone_rings)
            backbone_to_sidechain[backbone_atom] = sidechains
            beta_atoms: List[Atom] = self.get_beta_atoms(backbone, backbone_atom, backbone_rings)
            for sidechain in sidechains:
                beta_atom: Optional[Atom] = None

                for atom in sidechain:
                    if atom in beta_atoms:
                        beta_atom = atom
                        break

                assert beta_atom

                sidechain_bonds = self.get_sidechain_bonds(sidechain)
                masked_bond = self.structure.bond_lookup[beta_atom][backbone_atom]
                rotatable_bonds: List[Bond] = []

                for bond in sidechain_bonds:
                    if self.bond_is_rotatable(bond) and bond != masked_bond:
                        rotatable_bonds.append(bond)
                backbone_to_rotatable_bonds[backbone_atom] = rotatable_bonds


            if len(sidechains) == 2:
                pass
            elif len(sidechains) == 1:
                sidechain = sidechains[0]
                atoms_to_adjust: List[Atom] = self.find_high_atoms(sidechain, max_y)
                clashing_atoms = self.find_clashing_sidechain_atoms(sidechain)
                if clashing_atoms or atoms_to_adjust:
                    beta_atoms: List[Atom] = self.get_beta_atoms(backbone, backbone_atom, backbone_rings)
                    assert len(beta_atoms) == 1
                    beta_atom = beta_atoms[0]

                    sidechain_bonds = self.get_sidechain_bonds(sidechain)
                    rotatable_bonds: List[Bond] = []

                    masked_bond = self.structure.bond_lookup[beta_atom][backbone_atom]
                    for bond in sidechain_bonds:
                        if self.bond_is_rotatable(bond) and bond != masked_bond and \
                                (bond.atom_1 in beta_atoms or bond.atom_2 in beta_atoms):
                            rotatable_bonds.append(bond)

                    if not rotatable_bonds:
                        for bond in sidechain_bonds:
                            if self.bond_is_rotatable(bond) and bond != masked_bond:
                                rotatable_bonds.append(bond)

    def resolve_secondary_overlaps(self, sorted_scores: List[Tuple[float, Atom]]) -> None:
        for score, atom in sorted_scores:
            if score > self.options.overlap_sensitivity:
                if len(atom.drawn_neighbours) <= 1:
                    if atom.drawn_neighbours and atom.drawn_neighbours[0]._adjacent_to_stereobond():
                        continue

                    closest_atom = self.get_closest_atom(atom)

                    drawn_neighbours = closest_atom.drawn_neighbours

                    if len(drawn_neighbours) <= 1:
                        if not closest_atom.draw.previous_position:
                            closest_position = drawn_neighbours[0].draw.position
                        else:
                            closest_position = closest_atom.draw.previous_position

                    else:
                        if not closest_atom.draw.previous_position:
                            closest_position = drawn_neighbours[0].draw.position
                        else:
                            closest_position = closest_atom.draw.position

                    if not atom.draw.previous_position:
                        atom_previous_position = atom.drawn_neighbours[0].draw.position
                    else:
                        atom_previous_position = atom.draw.previous_position
                    #
                    # atom.draw.position.rotate_away_from_vector(closest_position, atom_previous_position,
                    #                                            math.radians(20))

    def rotate_subtree_ring(self, root, masked, angle, center):
        for atom in self.traverse_substructure(root, masked):
            atom.draw.position.rotate_around_vector(angle, center)
            for anchored_ring in atom.draw.anchored_rings:
                if anchored_ring.center:
                    anchored_ring.center.rotate_around_vector(angle, center)

    def get_beta_atoms(self, backbone: List[Atom], backbone_atom: Atom, backbone_rings: List[Atom]) -> List[Atom]:
        """
        Returns atoms directly adjacent to the backbone atom that is not in the same ring as a backbone atom

        Parameters
        ----------
        backbone: list of [Atom, ->] with each atom a backbone atom
        backbone_atom: Atom instance, backbone atom for which the sidechains are determined
        backbone_rings: list of [Atom, ->] with each atom a member of a ring that overlaps with the backbone

        Returns
        -------
        beta_atoms: list of [Atom, ->], representing the atoms directly adjacent to the backbone atom
        """
        beta_atoms: List[Atom] = []
        for neighbour in backbone_atom.neighbours:
            if neighbour.type != 'H' and not neighbour.annotations.domain_type:
                if neighbour not in backbone and neighbour not in backbone_rings:
                    beta_atoms.append(neighbour)

        return beta_atoms



    def get_sidechains(self, backbone: List[Atom], backbone_atom: Atom, backbone_rings: List[Atom]) -> List[List[Atom]]:
        """
        Returns the non-ring sidechains of a backbone atom

        Parameters
        ----------
        backbone: list of [Atom, ->] with each atom a backbone atom
        backbone_atom: Atom instance, backbone atom for which the sidechains are determined
        backbone_rings: list of [Atom, ->] with each atom a member of a ring that overlaps with the backbone

        Returns
        -------
        sidechains: list of [[Atom, ->] ->], representing the sidechains coming out of the backbone atom

        """
        sidechains: List[List[Atom]] = []
        beta_atoms = self.get_beta_atoms(backbone, backbone_atom, backbone_rings)
        for beta_atom in beta_atoms:
            sidechain: List[Atom] = []

            for atom in self.traverse_substructure(beta_atom, {backbone_atom}):
                sidechain.append(atom)
            sidechains.append(sidechain)
        return sidechains

    def get_sidechain_bonds(self, sidechain: List[Atom]) -> List[Bond]:
        """
        Returns a list of bonds that exist between sidechain atoms

        Parameters
        ----------
        sidechain: list of [Atom, ->]

        Returns
        -------
        sidechain_bonds: list of [Bond, ->]

        """
        bonds: Set[Bond] = set()

        for atom in sidechain:
            for neighbour in atom.neighbours:
                if neighbour in sidechain:
                    bonds.add(self.structure.bond_lookup[atom][neighbour])

        sidechain_bonds: List[Bond] = list(bonds)

        return sidechain_bonds

    def position_sidechains(self, backbone, backbone_to_placement, atoms_in_rings):
        backbone_atom_before = None
        backbone_to_neighbours = {}

        for backbone_atom in backbone:
            neighbours = []
            for neighbour in backbone_atom.neighbours:
                if neighbour.type != 'H' and not neighbour.annotations.domain_type:
                    if neighbour not in backbone and neighbour not in atoms_in_rings \
                            and neighbour not in neighbours:
                        neighbours.append(neighbour)
            backbone_to_neighbours[backbone_atom] = neighbours

        for backbone_atom, neighbours in backbone_to_neighbours.items():
            if backbone_atom in backbone_to_placement:
                placement = backbone_to_placement[backbone_atom]
                if len(neighbours) == 0:
                    pass
                elif len(neighbours) == 1:
                    neighbour = neighbours[0]

                    current_angle = get_angle(backbone_atom.draw.position, neighbour.draw.position)

                    if placement == 'right':
                        desired_angle = 180.0
                    elif placement == 'left':
                        desired_angle = 0.0
                    else:
                        raise ValueError("Placement must be 'left' or 'right'.")
                    # if ring in central chain but last member or tertiary carbon
                    if backbone_atom in atoms_in_rings and backbone_atom_before in atoms_in_rings and backbone_atom_before:
                        desired_angle += 180

                    required_rotation_deg = desired_angle - current_angle
                    required_rotation_rad = math.radians(required_rotation_deg)

                    self.rotate_subtree(neighbour, backbone_atom, required_rotation_rad, backbone_atom.draw.position)

                elif len(neighbours) == 2:
                    neighbour_1, neighbour_2 = neighbours

                    current_angle_1 = get_angle(backbone_atom.draw.position, neighbour_1.draw.position)
                    current_angle_2 = get_angle(backbone_atom.draw.position, neighbour_2.draw.position)

                    if placement == 'right':
                        desired_angle_1 = 140.0
                        desired_angle_2 = 220.0
                    elif placement == 'left':
                        desired_angle_1 = 40.0
                        desired_angle_2 = -40.0
                    else:
                        raise ValueError("Placement must be 'left' or 'right'.")

                    required_rotation_deg_1 = desired_angle_1 - current_angle_1
                    required_rotation_rad_1 = math.radians(required_rotation_deg_1)

                    required_rotation_deg_2 = desired_angle_2 - current_angle_2
                    required_rotation_rad_2 = math.radians(required_rotation_deg_2)

                    self.rotate_subtree(neighbour_1, backbone_atom, required_rotation_rad_1, backbone_atom.draw.position)
                    self.rotate_subtree(neighbour_2, backbone_atom, required_rotation_rad_2, backbone_atom.draw.position)
                elif len(neighbours) == 3:
                    if backbone_atom != backbone[-1]:
                        raise ValueError("Central backbone atoms can only have two non-backbone neighbours at most")
                    else:
                        neighbour_1, neighbour_2, neighbour_3 = neighbours

                        current_angle_1 = get_angle(backbone_atom.draw.position, neighbour_1.draw.position)
                        current_angle_2 = get_angle(backbone_atom.draw.position, neighbour_2.draw.position)
                        current_angle_3 = get_angle(backbone_atom.draw.position, neighbour_3.draw.position)

                        if placement == 'right':
                            desired_angle_1 = 140.0
                            desired_angle_2 = 220.0
                            desired_angle_3 = 60.0
                        elif placement == 'left':
                            desired_angle_1 = 40.0
                            desired_angle_2 = -40.0
                            desired_angle_3 = 120.0
                        else:
                            raise ValueError("Placement must be 'left' or 'right'.")

                        required_rotation_deg_1 = desired_angle_1 - current_angle_1
                        required_rotation_rad_1 = math.radians(required_rotation_deg_1)

                        required_rotation_deg_2 = desired_angle_2 - current_angle_2
                        required_rotation_rad_2 = math.radians(required_rotation_deg_2)

                        required_rotation_deg_3 = desired_angle_3 - current_angle_3
                        required_rotation_rad_3 = math.radians(required_rotation_deg_3)

                        self.rotate_subtree(neighbour_1, backbone_atom, required_rotation_rad_1,
                                            backbone_atom.draw.position)
                        self.rotate_subtree(neighbour_2, backbone_atom, required_rotation_rad_2,
                                            backbone_atom.draw.position)
                        self.rotate_subtree(neighbour_3, backbone_atom, required_rotation_rad_3,
                                            backbone_atom.draw.position)

                else:
                    raise ValueError("Terminal backbone atoms can only have three non-backbone neighbours at most")
                backbone_atom_before = backbone_atom
            # We are in a ring that we cannot linearise; rotate the whole ring straight down
            else:
                drawing_ring = self.get_ring(backbone_atom.draw.rings[0])
                center = Vector(0, 0)
                for atom in drawing_ring.members:
                    center.add(atom.draw.position)

                center.divide(len(drawing_ring.members))
                angle = get_angle(backbone_atom_before.draw.position,
                                  center)

                desired_angle = 90.0

                required_rotation_deg = desired_angle - angle
                required_rotation_rad = math.radians(required_rotation_deg)

                masked = {backbone_atom_before}

                # for atom in drawing_ring.members:
                for atom in self.traverse_substructure(backbone_atom, masked):

                    if atom != backbone_atom_before:
                        atom.draw.position.rotate_around_vector(required_rotation_rad, backbone_atom_before.draw.position)
                break

    def plot_halflines_s_domain(self, line, ax, midpoint):
        truncated_line = line.get_truncated_line(
            self.options.short_bond_length)
        self.plot_line_dashed(truncated_line, ax, color='#a6a6a6')

    def plot_line_dashed(self, line, ax, color='grey'):
        with matplotlib.rc_context({'path.sketch': (5, 10, 1)}):
            ax.plot([line.point_1.x, line.point_2.x],
                    [line.point_1.y, line.point_2.y], color=color,
                    linewidth=self.line_width/1.5)

    def get_delta_x_sidechain(self, sidechain_atom, neighbouring_carbon):
        """Returns the change tot the x-position of the sidegroup symbol in
        the structure drawing as float
        sidechain_atom: Atom object of the unknown sidechain group
        neighoburin_carbon: Atom object of the carbon in the central chain
        bound to the unknown sidechain group
        """
        sidechainatom_x = sidechain_atom.draw.position.x
        carbon_x = neighbouring_carbon.draw.position.x
        if (carbon_x - sidechainatom_x) < -10:
            delta_x = 1.5
        elif (carbon_x - sidechainatom_x) > 10:
            delta_x = -1.5
        else:
            delta_x = 0.0

        return delta_x

    def draw_svg(self, annotation: Union[None, str] = None, numbered_atoms: List = None) -> str:

        self.set_annotation_for_grouping(annotation)

        ring_centers_x = []
        ring_centers_y = []

        for ring in self.rings:
            self.set_ring_center(ring)

            ring_centers_x.append(ring.center.x)
            ring_centers_y.append(ring.center.y)

        for bond_nr, bond in self.structure.bonds.items():
            if bond.atom_1.draw.positioned and bond.atom_2.draw.positioned:
                line = Line(bond.atom_1.draw.position, bond.atom_2.draw.position, bond.atom_1, bond.atom_2)
                midpoint = line.get_midpoint()

                if bond.type == 'single':
                    if bond in self.chiral_bonds:

                        orientation, chiral_center = self.chiral_bond_to_orientation[bond]
                        self._draw_chiral_bond(orientation, chiral_center, line, midpoint)
                    else:
                        self._draw_halflines(line, midpoint)
                elif bond.type == 'double':
                    if not self._is_terminal(bond.atom_1) and not self._is_terminal(bond.atom_2):
                        self._draw_halflines(line, midpoint)

                        common_ring_numbers = self._get_common_rings(bond.atom_1, bond.atom_2)

                        if common_ring_numbers:
                            common_rings = []
                            for ring_nr in common_ring_numbers:
                                common_rings.append(self.get_ring(ring_nr))

                            common_rings.sort(key=lambda x: len(x.members))
                            common_ring = common_rings[0]
                            ring_centre = common_ring.center
                            second_line = line.double_line_towards_center(ring_centre, self.options.bond_spacing,
                                                                          self.options.double_bond_length)
                            second_line_midpoint = second_line.get_midpoint()
                            self._draw_halflines_double(second_line, second_line_midpoint)

                        else:
                            bond_neighbours = bond.atom_1.drawn_neighbours + bond.atom_2.drawn_neighbours
                            if bond_neighbours:
                                vectors = [atom.draw.position for atom in bond_neighbours]
                                gravitational_point = Vector.get_average(vectors)
                                second_line = line.double_line_towards_center(gravitational_point,
                                                                              self.options.bond_spacing,
                                                                              self.options.double_bond_length)
                                second_line_midpoint = second_line.get_midpoint()
                                self._draw_halflines_double(second_line, second_line_midpoint)
                            else:
                                print("Shouldn't happen!")
                    else:
                        if self._is_terminal(bond.atom_1) and self._is_terminal(bond.atom_2):
                            dummy_1 = Vector(bond.atom_1.draw.position.x + 1, bond.atom_1.draw.position.y + 1)
                            dummy_2 = Vector(bond.atom_1.draw.position.x - 1, bond.atom_1.draw.position.y - 1)
                            double_bond_line_1 = line.double_line_towards_center(dummy_1,
                                                                                 self.options.bond_spacing / 2.0,
                                                                                 self.options.double_bond_length)
                            double_bond_line_1_midpoint = double_bond_line_1.get_midpoint()
                            double_bond_line_2 = line.double_line_towards_center(dummy_2,
                                                                                 self.options.bond_spacing / 2.0,
                                                                                 self.options.double_bond_length)
                            double_bond_line_2_midpoint = double_bond_line_2.get_midpoint()

                            self._draw_halflines_double(double_bond_line_1, double_bond_line_1_midpoint)
                            self._draw_halflines_double(double_bond_line_2, double_bond_line_2_midpoint)

                        else:

                            if self._is_terminal(bond.atom_1):
                                terminal_atom = bond.atom_1
                                branched_atom = bond.atom_2
                            else:
                                terminal_atom = bond.atom_2
                                branched_atom = bond.atom_1

                            if len(branched_atom.drawn_neighbours) >= 3:
                                closest_two = self.get_sorted_distances_from_list(terminal_atom,
                                                                                  branched_atom.drawn_neighbours)
                                closest_atom_1 = closest_two[0][1]
                                closest_atom_2 = closest_two[1][1]

                                line = Line(terminal_atom.draw.position, branched_atom.draw.position, terminal_atom,
                                            branched_atom)

                                double_bond_line_1, double_bond_line_2 = line.get_perpendicular_lines(
                                    self.options.bond_spacing / 2.0)
                                terminal_atom_pos_1 = double_bond_line_1.get_atom_coords(terminal_atom)
                                terminal_atom_pos_2 = double_bond_line_2.get_atom_coords(terminal_atom)

                                closest_atom_to_pos_1 = terminal_atom_pos_1.get_closest_atom(closest_atom_1,
                                                                                             closest_atom_2)
                                closest_atom_to_pos_2 = terminal_atom_pos_2.get_closest_atom(closest_atom_1,
                                                                                             closest_atom_2)

                                bond_1_line = Line(branched_atom.draw.position, closest_atom_to_pos_1.draw.position,
                                                   branched_atom, closest_atom_to_pos_1)
                                bond_2_line = Line(branched_atom.draw.position, closest_atom_to_pos_2.draw.position,
                                                   branched_atom, closest_atom_to_pos_2)

                                double_bond_line_1_midpoint = double_bond_line_1.get_midpoint()
                                double_bond_line_2_midpoint = double_bond_line_2.get_midpoint()

                                intersection_1 = double_bond_line_1.find_intersection(bond_1_line)
                                intersection_2 = double_bond_line_2.find_intersection(bond_2_line)

                                if terminal_atom.draw.position.x > branched_atom.draw.position.x:
                                    # check for parallel lines
                                    if intersection_1 and intersection_1.x < 100000 and intersection_1.y < 100000:
                                        double_bond_line_1.point_1 = intersection_1
                                    if intersection_2 and intersection_2.x < 100000 and intersection_2.y < 100000:
                                        double_bond_line_2.point_1 = intersection_2

                                else:
                                    # check for parallel lines
                                    if intersection_1 and intersection_1.x < 100000 and intersection_1.y < 100000:
                                        double_bond_line_1.point_2 = intersection_1
                                    if intersection_2 and intersection_2.x < 100000 and intersection_2.y < 100000:
                                        double_bond_line_2.point_2 = intersection_2

                                self._draw_halflines(double_bond_line_1, double_bond_line_1_midpoint)
                                self._draw_halflines(double_bond_line_2, double_bond_line_2_midpoint)

                            else:
                                self._draw_halflines(line, midpoint)

                                bond_neighbours = bond.atom_1.drawn_neighbours + bond.atom_2.drawn_neighbours
                                if bond_neighbours:
                                    vectors = [atom.draw.position for atom in bond_neighbours]
                                    gravitational_point = Vector.get_average(vectors)
                                    second_line = line.get_parallel_line(gravitational_point,
                                                                         self.options.bond_spacing)
                                    second_line_midpoint = second_line.get_midpoint()
                                    self._draw_halflines(second_line, second_line_midpoint)
                                else:
                                    print("Shouldn't happen!")

                elif bond.type == 'triple':
                    self._draw_halflines(line, midpoint)
                    line_1, line_2 = line.get_parallel_lines(self.options.bond_spacing)
                    line_1_midpoint = line_1.get_midpoint()
                    line_2_midpoint = line_2.get_midpoint()
                    self._draw_halflines(line_1, line_1_midpoint)
                    self._draw_halflines(line_2, line_2_midpoint)

        for atom in self.structure.graph:
            if atom.draw.positioned:
                svg_text = ''
                svg_h_text = ''
                svg_charge_text = ''
                svg_h_count_text = ''

                if atom.type != 'C' or atom.draw.draw_explicit or atom.charge:
                    if atom.type == 'C' and not atom.charge:
                        svg_text = self._draw_text('.', atom.draw.position.x, atom.draw.position.y - 2,
                                                   color=atom.draw.colour)
                    else:
                        svg_text = self._draw_text(atom.type, atom.draw.position.x, atom.draw.position.y,
                                                   color=atom.draw.colour)
                if hasattr(atom, "domain_type") and atom.annotations.domain_type in ["Leader", "Follower", "ACP", "PCP"]:
                    svg_text = ''

                        # TODO: Make this possible in svg writing
                        # text = self.set_r_group_indices_subscript(atom.type)

                orientation = self._get_hydrogen_text_orientation(atom)

                # Swap up-down orientation due to swapped svg coordinate system
                if orientation == 'H_below_atom':
                    orientation = 'H_above_atom'
                elif orientation == 'H_above_atom':
                    orientation = 'H_below_atom'

                if atom.type != 'C' or atom.draw.draw_explicit or atom.charge:

                    hydrogen_count = 0

                    for neighbour in atom.neighbours:
                        if neighbour.type == 'H' and not neighbour.draw.is_drawn:
                            hydrogen_count += 1

                    h_x = atom.draw.position.x
                    h_y = atom.draw.position.y
                    h_subscript_x = atom.draw.position.x
                    h_subscript_y = atom.draw.position.y + 3
                    charge_x = atom.draw.position.x + 6
                    charge_y = atom.draw.position.y - 3

                    if orientation == 'H_below_atom':
                        h_y = atom.draw.position.y + 7
                        h_subscript_x += 2
                        h_subscript_y += 10
                    elif orientation == 'H_above_atom':
                        h_y = atom.draw.position.y - 7
                        h_subscript_x += 2
                        h_subscript_y -= 4
                    elif orientation == 'H_before_atom':
                        if hydrogen_count > 1:
                            h_x -= 10
                            h_subscript_x -= 5
                        elif hydrogen_count == 1:
                            h_x -= 6
                    else:
                        h_x += len(atom.type) * 6
                        h_subscript_x += len(atom.type) * 6 + 5

                    if abs(atom.charge):
                        if hydrogen_count and orientation == 'H_after_atom':
                            if hydrogen_count > 1:
                                charge_x += len(atom.type) * 6 + 7
                            else:
                                charge_x += len(atom.type) * 6 + 2

                    h_pos = Vector(h_x, h_y)
                    h_subscript_pos = Vector(h_subscript_x, h_subscript_y)
                    charge_pos = Vector(charge_x, charge_y)

                    if hydrogen_count:
                        svg_h_text = self._draw_text('H', h_pos.x, h_pos.y,
                                                     color=atom.draw.colour)
                        if hydrogen_count > 1:
                            svg_h_count_text = self._draw_text(hydrogen_count, h_subscript_pos.x, h_subscript_pos.y,
                                                               font_size=self.options.svg_font_size_small,
                                                               color=atom.draw.colour)
                    if atom.charge:
                        if atom.charge > 0:
                            charge_symbol = '+'
                        else:
                            charge_symbol = '-'

                        if abs(atom.charge) > 1:
                            charge_text = f"{abs(atom.charge)}{charge_symbol}"
                        else:
                            charge_text = charge_symbol

                        svg_charge_text = self._draw_text(charge_text, charge_pos.x, charge_pos.y,
                                                          font_size=self.options.svg_font_size_small,
                                                          color=atom.draw.colour)

                if svg_text or svg_h_text or svg_charge_text or svg_h_count_text:
                    if self.structure_id:
                        text_group = f'<g id="atom_{atom}_{self.structure_id}_text">\n'
                    else:
                        text_group = f'<g id="atom_{atom}_text">\n'
                    if svg_text:
                        text_group += svg_text
                        text_group += '\n'
                    if svg_h_text:
                        text_group += svg_h_text
                        text_group += '\n'
                    if svg_charge_text:
                        text_group += svg_charge_text
                        text_group += '\n'
                    if svg_h_count_text:
                        text_group += svg_h_count_text
                        text_group += '\n'
                    text_group += '</g>'
                    self._add_svg_element(text_group, atom)

        if numbered_atoms:
            self.show_atom_numbers(numbered_atoms)

        svg = self._assemble_svg()
        return svg


def get_angle(vector1, vector2):
    difference = Vector.subtract_vectors(vector1, vector2)
    difference_angle = difference.angle()
    return math.degrees(difference_angle)



ATTRIBUTES = ['in_central_chain', 'KR_ep_target', 'KR_red_target',
              'latest_elongation_o', 'latest_elongation_methyl', 'DH_target',
              'ER_target', 'domain_type', 'unknown_index', 'chiral_c_ep',
              'n_atom_nmeth', 'leaving_oh_o', 'leaving_oh_h', 'c2_acid',
              'terminal_c', 'terminal_o']


ENZYME_TO_NAME = {"METHYLTRANSFERASE": "methyltransferase",
                  "C_METHYLTRANSFERASE": "C-methyltransferase",
                  "N_METHYLTRANSFERASE": "N-methyltransferase",
                  "O_METHYLTRANSFERASE": "O-methyltransferase",
                  "HYDROXYLASE": "hydroxylation",
                  "EPOXIDASE": "epoxidation",
                  "PRENYLTRANSFERASE": "prenyltransferase",
                  "ACETYLTRANSFERASE": "acetyltransferase",
                  "ACYLTRANSFERASE": "acyltransferase",
                  "AMINOTRANSFERASE": "aminotransferase",
                  "HALOGENASE": "halogenase",
                  "DOUBLE_BOND_REDUCTASE": "double bond reduction",
                  "DOUBLE_BOND_ISOMERASE": "double bond shift",
                  "DEHYDROGENASE": "double bond formation",
                  "KETO_REDUCTASE": "ketoreduction",
                  "ALCOHOLE_DEHYDROGENASE": "alcohol dehydrogenase",
                  "PEPTIDASE": "peptidase",
                  "PROTEASE": "protease",
                  "MONOAMINE_OXIDASE": "monoamine oxidase",
                  "DEHYDRATASE": "dehydratase",
                  "THREONINE_SERINE_DEHYDRATASE": "dehydratase",
                  "DECARBOXYLASE": "decarboxylase",
                  "SPLICEASE": "splicease",
                  "ARGINASE": "arginase",
                  "OXIDATIVE_BOND_SYNTHASE": "oxidative bond formation",
                  "MACROLACTAM_SYNTHETASE": "macrolactam synthase",
                  "CYCLODEHYDRASE": "cyclodehydration",
                  "LANTHIPEPTIDE_CYCLASE": "lanthipeptide cyclase",
                  "LANTHIONINE_SYNTHETASE": "lanthionine synthetase",
                  "THIOPEPTIDE_CYCLASE": "thiopeptide cyclase",
                  "AMINO_ACID_EPIMERASE": "epimerase",
                  "HYDROLASE": "hydrolysis",
                  "REDUCTIVE_LYASE": "Reductive single bond breakage"}


ALL_PKS_ELONGATION_UNITS = ['malonylcoa', 'methylmalonylcoa',
                            'methoxymalonylacp', 'ethylmalonylcoa', 'pk']

AMINOACID_ONE_LETTER_TO_NAME = {"A": "alanine",
                                "R": "arginine",
                                "N": "asparagine",
                                "D": "aspartic acid",
                                "C": "cysteine",
                                "E": "glutamic acid",
                                "Q": "glutamine",
                                "G": "glycine",
                                "H": "histidine",
                                "I": "isoleucine",
                                "L": "leucine",
                                "K": "lysine",
                                "M": "methionine",
                                "F": "phenylalanine",
                                "P": "proline",
                                "S": "serine",
                                "T": "threonine",
                                "W": "tryptophan",
                                "Y": "tyrosine",
                                "V": "valine"}

AMINOACID_ONE_LETTER_TO_SMILES = {"A": "N[C@@]([H])(C)C(=O)", "C": "N[C@@]([H])(CS)C(=O)",
                                  "D": "N[C@@]([H])(CC(=O)O)C(=O)", "E": "N[C@@]([H])(CCC(=O)O)C(=O)",
                                  "F": "N[C@@]([H])(Cc1ccccc1)C(=O)", "G": "NCC(=O)", "H": "N[C@@]([H])(CC1=CN=CN1)C(=O)",
                                  "I": "N[C@@]([H])([C@]([H])(CC)C)C(=O)", "K": "N[C@@]([H])(CCCCN)C(=O)",
                                  "L": "N[C@@]([H])(CC(C)C)C(=O)", "M": "N[C@@]([H])(CCSC)C(=O)",
                                  "N": "N[C@@]([H])(CC(=O)N)C(=O)", "P": "N1[C@@]([H])(CCC1)C(=O)",
                                  "Q": "N[C@@]([H])(CCC(=O)N)C(=O)", "R": "N[C@@]([H])(CCCNC(=N)N)C(=O)",
                                  "S": "N[C@@]([H])(CO)C(=O)", "T": "N[C@@]([H])([C@]([H])(O)C)C(=O)",
                                  "V": "N[C@@]([H])(C(C)C)C(=O)", "W": "N[C@@]([H])(CC(=CN2)C1=C2C=CC=C1)C(=O)",
                                  "Y": "N[C@@]([H])(Cc1ccc(O)cc1)C(=O)"}

PRENYL_TRANSFERASE_SUBSTRATES_TO_SMILES = {
    "DIMETHYLALLYL": r"CC=C(C)C",
    "3_METHYL_1_BUTENYL": r"C(C)(C)C=C",
    "GERANYL": r"C\C=C(/C)CCC=C(C)C",
    "FARNESYL": r"C\C=C(/C)CC\C=C(/C)CCC=C(C)C",
    "GERANYLGERANYL": r"CC(C)=CCC\C(C)=C\CC\C=C(/C)CCC=C(C)C",
    "SQUALENE": r"CC(=CCC/C(=C/CC/C(=C/CC/C=C(/CC/C=C(/CCC=C(C)C)\C)\C)/C)/C)C",
    "PHYTOENE": r"CC(=CCC/C(=C/CC/C(=C/CC/C(=C/C=C\C=C(/C)\CC/C=C(\C)/CC/C=C(\C)/CCC=C(C)C)/C)/C)/C)C"
}

SUGARS_TO_SMILES = {

}



from dataclasses import dataclass
from typing import List, Union
import os


@dataclass
class MacrocyclizationRepresentation:
    atom_1: str
    atom_2: str
    type: str = "oxidative"


@dataclass
class CleavageSiteRepresentation:
    amino_acid: str
    amino_acid_index: int
    peptide_fragment_to_keep: str


@dataclass
class TailoringRepresentation:
    gene_name: str
    type: str
    # Some tailoring reactions involve more than one atom
    modification_sites: List[List[str]]
    substrate: Union[str, None] = None


@dataclass
class IsomerizationRepresentation:
    modification_sites: List[List[str]]


@dataclass
class MethylShiftRepresentation:
    modification_sites: List[List[str]]


@dataclass
class DomainRepresentation:
    gene_name: Union[str, None]
    type: str
    subtype: Union[str, None] = None
    name: Union[str, None] = None
    active: bool = True
    used: bool = True


@dataclass
class ModuleRepresentation:
    type: str
    subtype: Union[str, None]
    substrate: str
    domains: List[DomainRepresentation]
    iterations: int = 1


@dataclass
class ClusterRepresentation:
    modules: List[ModuleRepresentation]
    tailoring_enzymes: Union[List[TailoringRepresentation], None] = None

    def write_cluster(self, out_dir):
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)
        cluster_out = os.path.join(out_dir, "cluster.txt")
        with open(cluster_out, "w") as out:
            out.write(
                "gene_name\tmodule_nr\tmodule_type\tmodule_subtype\tmodule_substrate\tnr_iterations\tdomain_type\tdomain_subtype\tdomain_name\tdomain_used\tdomain_active\n"
            )
            for i, module in enumerate(self.modules):
                for domain in module.domains:
                    out.write(
                        f"{domain.gene_name}\t{i}\t{module.type}\t{module.subtype}\t{module.substrate}\t{module.iterations}\t{domain.type}\t{domain.subtype}\t{domain.name}\t{domain.used}\t{domain.active}\n"
                    )
        if self.tailoring_enzymes:
            tailoring_out = os.path.join(out_dir, "tailoring.txt")
            with open(tailoring_out, "w") as tailoring:
                tailoring.write("gene_name\ttype\tsubstrate\tmodification_sites\n")

                for enzyme in self.tailoring_enzymes:
                    if enzyme.modification_sites:
                        site_reprs = []
                        for modification_site in enzyme.modification_sites:
                            site_repr = "|".join(list(map(str, modification_site)))
                            site_reprs.append(site_repr)

                        site_str = ":".join(site_reprs)
                    else:
                        site_str = str(None)

                    tailoring.write(
                        f"{enzyme.gene_name}\t{enzyme.type}\t{enzyme.substrate}\t{site_str}\n"
                    )

    @classmethod
    def from_file(cls, in_dir):
        in_cluster = os.path.join(in_dir, "cluster.txt")
        in_tailoring = os.path.join(in_dir, "tailoring.txt")
        module_representations = []
        modules = {}
        with open(in_cluster, "r") as cluster:
            cluster.readline()
            for line in cluster:
                line = line.strip()
                if line:
                    line_info = line.split("\t")
                    line_info_cleaned = []
                    for i, entry in enumerate(line_info):
                        if line_info[i] == "False":
                            line_info_cleaned.append(False)
                        elif line_info[i] == "True":
                            line_info_cleaned.append(True)
                        elif line_info[i] == "None":
                            line_info_cleaned.append(None)
                        else:
                            line_info_cleaned.append(line_info[i])
                    (
                        gene_name,
                        module_nr,
                        module_type,
                        module_subtype,
                        substrate,
                        iterations,
                        domain_type,
                        domain_subtype,
                        domain_name,
                        domain_used,
                        domain_active,
                    ) = line_info_cleaned
                    module_nr = int(module_nr)
                    domain = DomainRepresentation(
                        gene_name,
                        domain_type,
                        domain_subtype,
                        domain_name,
                        domain_active,
                        domain_used,
                    )
                    if module_nr not in modules:
                        modules[module_nr] = {
                            "type": module_type,
                            "subtype": module_subtype,
                            "substrate": substrate,
                            "iterations": iterations,
                            "domains": [domain],
                        }
                    else:
                        modules[module_nr]["domains"].append(domain)

            for module_nr, module_data in modules.items():
                module = ModuleRepresentation(
                    module_data["type"],
                    module_data["subtype"],
                    module_data["substrate"],
                    module_data["domains"],
                    module_data["iterations"],
                )
                module_representations.append(module)

        if os.path.exists(in_tailoring):
            tailoring_enzymes = []
            with open(in_tailoring, "r") as tailoring:
                tailoring.readline()
                for line in tailoring:
                    line_info = line.split("\t")
                    line_info_cleaned = []
                    for i, entry in enumerate(line_info):
                        if line_info[i] == "None":
                            line_info_cleaned.append(None)
                        else:
                            line_info_cleaned.append(line_info[i])

                    gene_name, enzyme_type, substrate, site_str = line_info_cleaned

                    sites = []

                    if site_str:
                        site = site_str.split(":")
                        sites.append(site.split("|"))
                    tailoring_representation = TailoringRepresentation(
                        gene_name, enzyme_type, sites, substrate
                    )
                    tailoring_enzymes.append(tailoring_representation)

        else:
            tailoring_enzymes = None

        return cls(module_representations, tailoring_enzymes)




import os
from typing import Optional
from itertools import permutations

from sys import argv
from Bio import SeqIO
from raichu.representations import (
    ClusterRepresentation,
    ModuleRepresentation,
    DomainRepresentation,
)
from raichu.run_raichu import draw_cluster
from raichu.substrate import PksStarterSubstrate
from dataclasses import dataclass, field


NRPS_PKS_RULES = {
    "biosynthetic-additional (rule-based-clusters) AMP-binding",
    "biosynthetic-additional (rule-based-clusters) PP-binding",
    "biosynthetic (rule-based-clusters) NRPS: Condensation",
    "biosynthetic (rule-based-clusters) T1PKS: PKS_AT",
    "biosynthetic (rule-based-clusters) T1PKS: mod_KS",
}

AS_TO_NRPS = {
    "ala": "alanine",
    "arg": "arginine",
    "asn": "asparagine",
    "asp": "aspartic acid",
    "cys": "cysteine",
    "gln": "glutamine",
    "glu": "glutamic acid",
    "gly": "glycine",
    "his": "histidine",
    "ile": "isoleucine",
    "leu": "leucine",
    "lys": "lysine",
    "met": "methionine",
    "phe": "phenylalanine",
    "pro": "proline",
    "ser": "serine",
    "thr": "threonine",
    "trp": "tryptophan",
    "tyr": "tyrosine",
    "val": "valine",
    "3-me-glu": "4-methylglutamic acid",
    "4ppro": "**Unknown**",
    "aad": "2-aminoadipic acid",
    "abu": "2-aminobutyric acid",
    "aeo": "2-amino-9,10-epoxy-8-oxodecanoid acid",
    "ala-b": "beta-alanine",
    "ala-d": "d-alanine",
    "allo-thr": "allo-threonine",
    "b-ala": "beta-alanine",
    "beta-ala": "beta-alanine",
    "bmt": "4-butenyl-4-methyl threonine",
    "cap": "capreomycidine",
    "bht": "**Unknown**",
    "dab": "2,4-diaminobutyric acid",
    "dhb": "2,3-dihydroxybenzoic acid",
    "dhpg": "3,5-dihydroxyphenylglycine",
    "dht": "dehydrobutyrine",
    "dpg": "3,5-dihydroxyphenylglycine",
    "hiv": "2-hydroxyisovalerate",
    "hiv-d": "d-2-hydroxyisovalerate",
    "hmp-d": "**Unknown**",
    "horn": "**Unknown**",
    "hpg": "4-hydroxyphenylglycine",
    "hyv": "4-hydroxyvaline",
    "hyv-d": "**Unknown**",
    "iva": "isovaline",
    "lys-b": "beta-lysine",
    "orn": "ornithine",
    "phg": "phenylglycine",
    "pip": "pipecolic acid",
    "sal": "salicylic acid",
    "tcl": "**Unknown**",
    "vol": "valinol",
    "ldap": "**Unknown**",
    "meval": "tert-leu",
    "alle": "allo-isoleucine",
    "alaninol": "alaninol",
    "n-(1,1-dimethyl-1-allyl)trp": "**Unknown**",
    "d-lyserg": "d-lysergic acid",
    "ser-thr": "**Unknown**",
    "mephe": "**Unknown**",
    "haorn": "**Unknown**",
    "hasn": "**Unknown**",
    "hforn": "**Unknown**",
    "s-nmethoxy-trp": "**Unknown**",
    "alpha-hydroxy-isocaproic-acid": "**Unknown**",
    "mehoval": "**Unknown**",
    "2-oxo-isovaleric-acid": "alpha-ketoisovaleric acid",
    "aoda": "**Unknown**",
    "x": "**Unknown**",
}

antiSMASH_DOMAIN_TO_RAICHU_DOMAIN = {
    "PKS_KS": "KS",
    "PKS_AT": "AT",
    "PKS_PP": "CP",
    "PCP": "CP",
    "ACP": "CP",
    "PKS_KR": "KR",
    "PKS_DH": "DH",
    "PKS_DH2": "DH",
    "PKS_DHt": "DH",
    "PKS_ER": "ER",
    "Thioesterase": "TE",
    "AMP-binding": "A",
    "Epimerization": "E",
    "Condensation": "C",
    "nMT": "nMT",
    "PP-binding": "CP",
    "Heterocyclization": "CYC",
    "CAL_domain": "CAL",
    "TD": "TD"
}

domains_to_be_refined = ["KR", "KS", "A", "AT"]

AS_TO_PKS = {
    "mal": "MALONYL_COA",
    "mmal": "METHYLMALONYL_COA",
    "mxmal": "METHOXYMALONYL_COA",
    "emal": "ETHYLMALONYL_COA",
}

DOMAIN_TO_MODULE_TYPE = {"KS": "PKS",
                         "AT": "PKS",
                         "ER": "PKS",
                         "DH": "PKS",
                         "KR": "PKS",
                         "C": "NRPS",
                         "A": "NRPS",
                         "E": "NRPS",
                         "nMT": "NRPS",
                         "CYC": "NRPS",
                         "CP": "EITHER",
                         "TE": "EITHER",
                         "TD": "EITHER",
                         "CAL": "EITHER",
                         "UNKNOWN": "EITHER"}


@dataclass
class AntiSmashDomain:
    id: str
    type: str
    start: int
    end: int
    strand: str
    gene: str

    subtype: Optional[str] = None
    active: bool = True
    substrate: Optional[str] = None
    name: Optional[str] = None
    raichu_type: Optional[str] = None
    raichu_subtype: Optional[str] = None

    def set_raichu_type(self):
        if self.type in antiSMASH_DOMAIN_TO_RAICHU_DOMAIN:
            self.raichu_type = antiSMASH_DOMAIN_TO_RAICHU_DOMAIN[self.type]
        elif self.type == 'MT' and self.subtype and self.subtype == 'nMT':
            self.raichu_type = 'nMT'
        else:
            self.raichu_type = "UNKNOWN"
            self.name = self.type

    def add_domain_information(self, features):
        if "domain_subtypes" in features:
            self.subtype = features["domain_subtypes"][0]
        if "specificity" in features:
            for spec in features["specificity"]:
                if "consensus:" in spec:
                    specificity = spec.split("consensus:")[1].strip().lower()

                    if specificity in AS_TO_NRPS:
                        substrate_name = AS_TO_NRPS[specificity]
                        self.substrate = substrate_name
                    elif specificity in AS_TO_PKS:
                        substrate_name = AS_TO_PKS[specificity]
                        self.substrate = substrate_name
                elif "KR activity:" in spec:
                    self.active = True if spec.split("KR activity:")[1].strip() == "active" else False

                elif "KR stereochemistry:" in spec:
                    domain_subtype = spec.split("KR stereochemistry:")[1].strip()
                    if domain_subtype == '(unknown)':
                        self.raichu_subtype = "UNKNOWN"
                    else:
                        self.raichu_subtype = domain_subtype

                elif "transATor:" in spec:
                    self.raichu_subtype = spec.split("transATor:")[1].strip().replace("-", "_").replace("/", "_").upper().replace("(UNKNOWN)", "MISCELLANEOUS")
        if not self.active and self.raichu_subtype == "C2":
            self.active = True
            
    def __repr__(self):
        return f"{self.raichu_type}_{self.start}_{self.end}"


@dataclass
class AntiSmashModule:
    start: int
    end: int
    strand: int

    complete: bool = False
    starter_complete: bool = False
    type: Optional[str] = None
    subtype: Optional[str] = None
    domains: list[AntiSmashDomain] = field(default_factory=list)

    def __post_init__(self):
        self.set_module_type()
        self.determine_completeness()

    def __eq__(self, other):
        if self.start == other.start and self.end == other.end and self.strand == other.strand:
            return True
        return False

    def __hash__(self):
        return self.start, self.end

    def __repr__(self):
        return f"{self.start}_{self.end}_{'-'.join([d.raichu_type for d in self.domains])}"

    def set_module_type(self):

        for domain in self.domains:
            if domain.raichu_type == 'C':
                if self.type == 'NRPS' or self.type is None:
                    self.type = 'NRPS'
                else:
                    raise ValueError(f"Can't set module type for module {self.__repr__()}. Cannot have C domain in {self.type} module.")

            if domain.raichu_type == 'CYC':
                if self.type == 'NRPS' or self.type is None:
                    self.type = 'NRPS'
                else:
                    raise ValueError(f"Can't set module type for module {self.__repr__()}. Cannot have heterocyclistion domain in {self.type} module.")

            if domain.raichu_type == 'A':
                if self.type == 'NRPS' or self.type is None:
                    self.type = 'NRPS'
                else:
                    raise ValueError(
                        f"Can't set module type for module {self.__repr__()}. Cannot have A domain in {self.type} module.")
            if domain.raichu_type == 'KR':
                if self.type == 'PKS' or self.type is None:
                    self.type = 'PKS'
                else:
                    raise ValueError(
                        f"Can't set module type for module {self.__repr__()}. Cannot have KR domain in {self.type} module.")

            if domain.raichu_type == 'KS':
                if self.type == 'PKS' or self.type is None:
                    self.type = 'PKS'
                else:
                    raise ValueError(
                        f"Can't set module type for module {self.__repr__()}. Cannot have KS domain in {self.type} module.")

            if domain.raichu_type == 'ER':
                if self.type == 'PKS' or self.type is None:
                    self.type = 'PKS'
                else:
                    raise ValueError(
                        f"Can't set module type for module {self.__repr__()}. Cannot have ER domain in {self.type} module.")

            if domain.raichu_type == 'DH':
                if self.type == 'PKS' or self.type is None:
                    self.type = 'PKS'
                else:
                    raise ValueError(
                        f"Can't set module type for module {self.__repr__()}. Cannot have DH domain in {self.type} module.")

            if domain.raichu_type == 'AT':

                if self.type == 'PKS' or self.type is None:
                    self.type = 'PKS'
                else:
                    raise ValueError(
                        f"Can't set module type for module {self.__repr__()}. Cannot have AT domain in {self.type} module.")

        if self.type == 'PKS':
            has_recognition_domain = False
            for domain in self.domains:
                if domain.raichu_type == 'AT' or domain.raichu_type == 'CAL':
                    has_recognition_domain = True

            if has_recognition_domain:
                self.subtype = 'PKS_CIS'
            else:
                self.subtype = 'PKS_TRANS'

        if self.type is None:
            self.type = "NRPS"

    def determine_completeness(self):
        self.complete = False
        self.starter_complete = False

        if self.type == 'PKS':

            has_ks = False
            has_acp = False
            has_at = False
            has_cal = False

            for domain in self.domains:
                if domain.raichu_type == 'KS':
                    has_ks = True
                if domain.raichu_type == 'CP':
                    has_acp = True
                if domain.raichu_type == 'AT':
                    has_at = True
                if domain.raichu_type == 'CAL':
                    has_cal = True

            if has_acp:
                if self.subtype == 'PKS_TRANS':
                    self.starter_complete = True
                elif self.subtype == 'PKS_CIS' and (has_at or has_cal):
                    self.starter_complete = True

                if has_ks:
                    if self.subtype == 'PKS_TRANS':
                        self.complete = True
                    elif self.subtype == 'PKS_CIS' and has_at:
                        self.complete = True

        elif self.type == 'NRPS':
            has_c = False
            has_cyc = False
            has_a = False
            has_pcp = False
            has_cal = False

            for domain in self.domains:
                if domain.raichu_type == 'C':
                    has_c = True
                if domain.raichu_type == 'CP':
                    has_pcp = True
                if domain.raichu_type == 'A':
                    has_a = True
                if domain.raichu_type == 'CAL':
                    has_cal = True
                if domain.raichu_type == 'CYC':
                    has_cyc = True

            if has_pcp and (has_a or has_cal):
                self.starter_complete = True

                if has_c or has_cyc:
                    self.complete = True

    def is_broken(self, module_nr):
        if module_nr == 0:
            if self.complete or self.starter_complete:
                return False
        elif len(self.domains) == 1 and self.domains[0].raichu_type == 'UNKNOWN':
            return False

        elif self.complete:
            return False

        return True

    def has_termination_domain(self):
        for domain in self.domains:
            if domain.raichu_type in ["TE", "TD"]:
                return True
        return False

    def has_c_starter(self):
        if self.domains and self.domains[0].subtype == "Condensation_Starter":
            return True
        return False

    def is_potential_starter(self):
        if self.starter_complete and not self.complete:
            return True

        return False


@dataclass
class ModuleBlock:
    start: int
    end: int
    strand: int
    modules: list[AntiSmashModule] = field(default_factory=list)
    is_starter_candidate: bool = False
    has_c_starter: bool = False
    has_termination_domain: bool = False

    def __post_init__(self):
        self.label_module_block()

    def __repr__(self):
        return '\n'.join([m.__repr__() for m in self.modules])

    def __eq__(self, other):
        if self.start == other.start and self.end == other.end and self.strand == other.strand:
            return True

        return False

    def __hash__(self):
        return self.start, self.end, self.strand

    def label_module_block(self):
        if self.modules:
            if self.modules[0].is_potential_starter():
                self.is_starter_candidate = True
            if self.modules[0].has_c_starter():
                self.has_c_starter = True
            if self.modules[-1].has_termination_domain():
                self.has_termination_domain = True


class ModuleOrder:

    def __init__(self, module_blocks: list[ModuleBlock]):
        self.module_blocks = module_blocks
        self.block_order = module_blocks
        self.modules_broken = None
        self.modules = []
        self.standalone_modules = []

    def sort_collinear_blocks(self):

        starter_candidates = []
        terminator_candidates = []

        for block in self.module_blocks:
            if block.is_starter_candidate:
                starter_candidates.append(block)

        for block in self.module_blocks:
            if block.has_c_starter:
                starter_candidates.append(block)
            if block.has_termination_domain:
                terminator_candidates.append(block)

        found_optimal_solution = False
        best_block_order = self.block_order[:]
        best_solution = self.check_broken_modules()

        if starter_candidates and terminator_candidates:
            for starter_candidate in starter_candidates:
                for terminator_candidate in terminator_candidates:
                    central_blocks = []
                    for module_block in self.module_blocks:
                        if module_block != starter_candidate and module_block != terminator_candidate:
                            central_blocks.append(module_block)

                    for module_order in permutations(central_blocks):
                        self.block_order = [starter_candidate] + list(module_order) + [terminator_candidate]
                        nr_broken_modules = self.check_broken_modules()
                        if nr_broken_modules < best_solution:
                            best_block_order = self.block_order[:]
                            best_solution = nr_broken_modules
                        if nr_broken_modules == 0:
                            found_optimal_solution = True
                            break

                    if found_optimal_solution:
                        break
                if found_optimal_solution:
                    break
        elif starter_candidates:
            for starter_candidate in starter_candidates:
                central_blocks = []
                for module_block in self.module_blocks:
                    if module_block != starter_candidate:
                        central_blocks.append(module_block)

                for module_order in permutations(central_blocks):
                    self.block_order = [starter_candidate] + list(module_order)
                    nr_broken_modules = self.check_broken_modules()
                    if nr_broken_modules < best_solution:
                        best_block_order = self.block_order[:]
                        best_solution = nr_broken_modules
                    if nr_broken_modules == 0:
                        found_optimal_solution = True
                        break
                if found_optimal_solution:
                    break

        elif terminator_candidates:
            for terminator_candidate in terminator_candidates:
                central_blocks = []
                for module_block in self.module_blocks:
                    if module_block != terminator_candidate:
                        central_blocks.append(module_block)

                for module_order in permutations(central_blocks):
                    self.block_order = list(module_order) + [terminator_candidate]
                    nr_broken_modules = self.check_broken_modules()
                    if nr_broken_modules < best_solution:
                        best_block_order = self.block_order[:]
                        best_solution = nr_broken_modules
                    if nr_broken_modules == 0:
                        found_optimal_solution = True
                        break
                if found_optimal_solution:
                    break

        self.block_order = best_block_order
        self.modules_broken = self.check_broken_modules()

        self.move_broken_modules()

    def remove_standalone_domains(self):

        for module_block in self.module_blocks[:]:
            if len(module_block.modules) == 1 and len(module_block.modules[0].domains) == 1 and \
                    module_block.modules[0].domains[0].raichu_type not in ["TE", "TD"]:
                self.module_blocks.remove(module_block)
                self.standalone_modules.append(module_block.modules[0])

    def move_broken_modules(self):
        module_block = self.recreate_modules()
        modules = module_block.modules[:]
        broken_modules = []
        complete_modules = []
        end_cluster = False

        for i, module in enumerate(module_block.modules):
            if module.is_broken(i) or end_cluster:

                broken_modules.append(module)
            else:
                complete_modules.append(module)

            if module.has_termination_domain() and not module.is_broken(i):
                end_cluster = True

        genes_with_complete_modules = set()

        for module in complete_modules:
            for domain in module.domains:
                genes_with_complete_modules.add(domain.gene)

        movable_broken_modules = []

        for broken_module in broken_modules:
            genes = set()
            for domain in broken_module.domains:
                genes.add(domain.gene)

            if not genes.intersection(genes_with_complete_modules):
                movable_broken_modules.append(broken_module)

        gene_to_broken = {}

        for module in movable_broken_modules:
            for domain in module.domains:
                if domain.gene not in gene_to_broken:
                    gene_to_broken[domain.gene] = []
                if module not in gene_to_broken[domain.gene]:
                    gene_to_broken[domain.gene].append(module)

        for gene, gene_modules in gene_to_broken.items():
            if gene_modules:
                modules_start = min([m.start for m in gene_modules] + [m.end for m in gene_modules])

                for gene_module in gene_modules:
                    modules.remove(gene_module)

                inserted = False
                for i, module in enumerate(modules[:]):
                    if module.start > modules_start:
                        modules = modules[:i] + gene_modules + modules[i:]
                        inserted = True
                        break

                if not inserted:
                    modules += gene_modules

        for standalone_module in self.standalone_modules:
            inserted = False
            for i, module in enumerate(modules[:]):
                if module.start > standalone_module.start:
                    modules = modules[:i] + [standalone_module] + modules[i:]
                    inserted = True
                    break

            if not inserted:
                modules.append(standalone_module)

        self.modules = modules

    def make_raichu_cluster(self):
        module_representations = []
        for i, module in enumerate(self.modules):
            domain_representations = []

            substrates = [domain.substrate for domain in module.domains if domain.substrate is not None]
            if len(substrates) > 0:
                substrate = substrates[0]

                if module.type == "PKS" and i == 0:
                    if substrate not in [v.name for v in PksStarterSubstrate]:
                        substrate = "WILDCARD"
            else:
                if module.type == "NRPS":
                    substrate = "**Unknown**"
                elif module.type == "PKS":
                    substrate = "WILDCARD"
                else:
                    raise ValueError("Module type can only be NRPS or PKS")

            for domain in module.domains:
                if domain.raichu_type == 'CP':
                    if module.type == 'NRPS':
                        domain.raichu_type = "PCP"
                    else:
                        domain.raichu_type = "ACP"

                if module.type == "PKS":
                    if domain.raichu_type == "CAL" and substrate == "WILDCARD":
                        substrate = "**Unknown**"

                domain_representation = DomainRepresentation(domain.gene, domain.raichu_type, domain.raichu_subtype,
                                                             domain.active, True)
                if domain_representation.type in [d.type for d in domain_representations]:
                    domain_representation.used = False
                if domain_representation.type == 'UNKNOWN':
                    domain_representation.used = False
                if domain_representation.type in ["AT", "A", "CAL"]:
                    if ["AT"] in [d.type for d in domain_representations] or \
                            ["A"] in [d.type for d in domain_representations] or \
                            ["CAL"] in [d.type for d in domain_representations]:
                        domain_representation.used = False

                domain_representations.append(domain_representation)
            module_representation = ModuleRepresentation(module.type, module.subtype, substrate, domain_representations)
            module_representations.append(module_representation)
        cluster_representation = ClusterRepresentation(module_representations)
        return cluster_representation

    def recreate_modules(self):
        domains = []
        for block in self.block_order:
            for module in block.modules:
                domains += module.domains

        module_block = make_modules(domains)
        return module_block

    def check_broken_modules(self):
        module_block = self.recreate_modules()
        nr_broken = 0
        end_cluster = False

        for i, module in enumerate(module_block.modules):
            if end_cluster:
                nr_broken += 1
                continue
            if module.is_broken(i):
                nr_broken += 1
            if module.has_termination_domain() and not module.is_broken(i):
                end_cluster = True

        return nr_broken


@dataclass
class AntiSmashGene:
    name: str
    start: int
    end: int
    strand: int
    synonyms: list[str] = field(default_factory=list)
    domains: list[AntiSmashDomain] = field(default_factory=list)
    modules: list[AntiSmashModule] = field(default_factory=list)

    def sort_domains(self):
        self.domains.sort(key=lambda d: d.start)
        if self.strand == -1:
            self.domains.reverse()

    def sort_modules(self):
        self.modules.sort(key=lambda m: m.start)
        if self.strand == -1:
            self.modules.reverse()

        for i, module in enumerate(self.modules):
            module.nr = i

    def __eq__(self, other):
        if self.start == other.start and self.end == other.end and self.strand == other.strand:
            return True
        return False

    def __hash__(self):
        return self.start, self.end

    def __repr__(self):
        return f"{self.name}"


def get_nrps_pks_modules(antismash_gbk):
    as_domains = parse_antismash_domains_gbk(antismash_gbk)
    genes = []
    for record in SeqIO.parse(antismash_gbk, "genbank"):
        for feature in record.features:
            if feature.type == 'CDS':

                qualifiers = feature.qualifiers
                synonyms = []

                if 'gene' in qualifiers:
                    synonyms.append(qualifiers['gene'][0])
                if 'protein' in qualifiers:
                    synonyms.append(qualifiers['protein'][0])
                if 'locus_tag' in qualifiers:
                    synonyms.append(qualifiers['locus_tag'][0])
                if 'protein_id' in qualifiers:
                    synonyms.append(qualifiers['protein_id'][0])

                if synonyms:
                    gene = AntiSmashGene(synonyms[0],
                                         int(feature.location.start),
                                         int(feature.location.end),
                                         int(feature.location.strand),
                                         synonyms)
                    if gene not in genes:
                        genes.append(gene)
                else:
                    print(f"Warning: No identifier found for a CDS. Domains on this gene will not be processed.")
                    # for domain in as_domains:
                    #     if domain.gene in synonyms:

    for domain in as_domains:
        gene_found = False
        for gene in genes:
            if domain.gene in gene.synonyms:
                gene_found = True
                gene.domains.append(domain)
                break
        if not gene_found:
            print(f"Warning: No CDS found for {domain.gene}. Domains on this gene will not be processed.")

    for gene in genes:
        gene.sort_domains()

    gene_groups = sort_genes(genes)

    filtered_groups = []
    for gene_group in gene_groups:
        filtered_groups += filter_genes(gene_group)

    modules_per_collinear_block = ModuleOrder(make_modules_from_gene_groups(filtered_groups))
    if len(filtered_groups) >= 6:
        modules_per_collinear_block.remove_standalone_domains()
    modules_per_collinear_block.sort_collinear_blocks()

    module_blocks = modules_per_collinear_block

    # If there are broken modules, also play around with the gene order
    #
    # if modules_per_collinear_block.modules_broken:
    #     filtered_genes = []
    #
    #     for gene_group in filtered_groups:
    #         for gene in gene_group:
    #             filtered_genes.append([gene])
    #
    #     if len(filtered_genes) < 5:
    #
    #         modules_per_gene = ModuleOrder(make_modules_from_gene_groups(filtered_genes))
    #         modules_per_gene.sort_collinear_blocks()
    #
    #         if modules_per_collinear_block.modules_broken > modules_per_gene.modules_broken:
    #             module_blocks = modules_per_gene

    return module_blocks


def make_modules_from_gene_groups(gene_groups):
    module_groups = []
    for gene_group in gene_groups:
        domains = []

        for gene in gene_group:
            domains += gene.domains

        module_block = make_modules(domains)

        module_groups.append(module_block)

    if not module_groups:
        raise Exception("Cluster is empty. This can happen with Type III PKS clusters. Please check the input file.")

    return module_groups


def make_modules(domains):

    modules = []
    module = []
    end_cluster = False

    type_i_te_domains = []
    type_ii_te_domains = []

    for i, domain in enumerate(domains):
        if domain.raichu_type in ["TD", "TE"]:
            if i == 0:
                type_ii_te_domains.append(domain)
            else:
                if domains[i - 1].gene == domain.gene:
                    type_i_te_domains.append(domain)
                else:
                    type_ii_te_domains.append(domain)

    for i, domain in enumerate(domains):
        if not module and not modules:
            module.append(domain)
            if domain.raichu_type in ["TD", "TE"]:
                modules.append(module)
                module = []
                if not type_i_te_domains:
                    end_cluster = True

        # When encountering a synthesis domain, start a new module

        elif (DOMAIN_TO_MODULE_TYPE[domain.raichu_type] == 'PKS' and "NRPS" in [DOMAIN_TO_MODULE_TYPE[d.raichu_type] for d in module]) or \
                (DOMAIN_TO_MODULE_TYPE[domain.raichu_type] == 'NRPS' and "PKS" in [DOMAIN_TO_MODULE_TYPE[d.raichu_type] for d in module]):
            if module:
                modules.append(module)
                module = [domain]
            else:
                raise Exception("Module can't be empty if it contains a domain. Fix code.")

        elif domain.raichu_type in ["C", "KS", "CYC"]:
            if module:
                # Don't start a new module when the previous domain could be a docking domain
                if module[-1].gene == domain.gene and module[-1].raichu_type == "UNKNOWN" and len(module) == 1:
                    module.append(domain)
                else:
                    modules.append(module)
                    module = [domain]
            else:
                module = [domain]

        elif domain.raichu_type in ["TD", "TE"]:
            if not end_cluster:
                if 'CP' in [d.raichu_type for d in module]:
                    if domain in type_i_te_domains or not type_i_te_domains:
                        module.append(domain)
                        modules.append(module)
                        module = []
                    else:
                        if module:
                            modules.append(module)

                        module = [domain]
                        modules.append(module)
                        module = []

                # Dealing with a standalone TE domain
                else:
                    if module:
                        modules.append(module)

                    module = [domain]
                    modules.append(module)
                    module = []

            # If a previous termination domain has been encountered, always consider the domain as a standalone domain

            else:
                if module:
                    modules.append(module)
                module = [domain]
                modules.append(module)
                module = []

            if domain in type_i_te_domains:
                end_cluster = True
            elif not type_i_te_domains:
                end_cluster = True

        # If a domain of known function is encountered already in the active module
        elif (domain.raichu_type in [d.raichu_type for d in module] and domain.raichu_type != "UNKNOWN") or \
                (domain.raichu_type in ["AT", "A", "CAL"] and ("AT" in [d.raichu_type for d in module] or
                                                               "A" in [d.raichu_type for d in module] or
                                                               "CAL" in [d.raichu_type for d in module])):
            if i != 0:
                previous_domain = domains[i - 1]
                # If the domain is on a new gene, always start a new module
                if previous_domain.gene != domain.gene:
                    modules.append(module)
                    module = [domain]

                # If the domain is on the same gene, start a new module if a carrier domain has already been encountered
                else:

                    # Double CP domains in a module CAN occur, so don't start a new module when the duplicated domain
                    # Is a CP domain on the same gene

                    if 'CP' in [d.raichu_type for d in module] and domain.raichu_type != 'CP':
                        if module:
                            modules.append(module)
                        module = [domain]
                    else:
                        module.append(domain)

            else:
                raise Exception("Module can't be empty if it contains a domain. Fix code.")

        elif domain.raichu_type == "UNKNOWN":
            if i != 0:
                previous_domain = domains[i - 1]
                if previous_domain.gene != domain.gene:
                    if module:
                        modules.append(module)

                    module = [domain]

                    if i != len(domains) - 1:
                        next_domain = domains[i + 1]
                        if next_domain.gene != domain.gene:
                            modules.append(module)
                            module = []

                else:
                    module.append(domain)

            else:
                raise Exception("Module can't be empty if it contains a domain. Fix code.")

        else:
            module.append(domain)

    if module:
        modules.append(module)

    antismash_modules = []

    for module in modules:
        start = min([domain.start for domain in module] + [domain.end for domain in module])
        end = max([domain.start for domain in module] + [domain.end for domain in module])
        strand = module[0].strand
        antismash_module = AntiSmashModule(start, end, strand, domains=module)
        antismash_modules.append(antismash_module)

    if antismash_modules:

        block_start = min([module.start for module in antismash_modules] + [module.end for module in antismash_modules])
        block_end = max([module.start for module in antismash_modules] + [module.end for module in antismash_modules])
        block_strand = antismash_modules[0].strand
        module_block = ModuleBlock(block_start, block_end, block_strand, antismash_modules)
    else:
        raise Exception("Cluster is empty.This can happen with Type III PKS clusters. Please check the input file.")

    return module_block


def filter_genes(genes):
    gene_groups = []
    gene_group = []

    for gene in genes:

        if gene.domains:
            gene_group.append(gene)
        else:
            if gene_group:
                gene_groups.append(gene_group)
                gene_group = []

    if gene_group:
        gene_groups.append(gene_group)

    return gene_groups


def sort_genes(genes):
    genes.sort(key=lambda gene: gene.start)
    gene_groups = []
    gene_group = []
    previous_direction = None

    for i, gene in enumerate(genes):
        current_direction = gene.strand
        if current_direction != previous_direction:
            if gene_group:
                if previous_direction == -1:
                    gene_group.reverse()
                gene_groups.append(gene_group)

            gene_group = [gene]
            previous_direction = current_direction

        else:
            gene_group.append(gene)

        if i == len(genes) - 1:
            if current_direction == -1:
                gene_group.reverse()
            gene_groups.append(gene_group)

    return gene_groups


def parse_antismash_domains_gbk(antismash_gbk, version="7.1.0"):
    domains = []

    for record in SeqIO.parse(antismash_gbk, "genbank"):
        for feature in record.features:
            if feature.type == "aSDomain":
                domain = AntiSmashDomain(feature.qualifiers["domain_id"][0],
                                         feature.qualifiers["aSDomain"][0],
                                         int(feature.location.start),
                                         int(feature.location.end),
                                         feature.location.strand,
                                         feature.qualifiers["locus_tag"][0])
                domain.add_domain_information(feature.qualifiers)
                domain.set_raichu_type()
                domains.append(domain)

    return domains


def load_antismash_gbk(gbk_file, version=7.0):
    module_blocks = get_nrps_pks_modules(gbk_file)
    cluster_representation = module_blocks.make_raichu_cluster()
    return cluster_representation


def parse_antismash_to_cluster_file(gbk_file, out_directory=None, version=7.0):
    if not out_directory:
        out_directory = os.path.splitext(gbk_file)[0]
    if not os.path.exists(out_directory):
        os.mkdir(out_directory)
    cluster = load_antismash_gbk(gbk_file)
    cluster.write_cluster(out_directory)


if __name__ == "__main__":
    draw_cluster(load_antismash_gbk(argv[1]), "test.svg")


from raichu.smiles_handling import get_smiles
from raichu.data.molecular_moieties import make_elongation_monomer, make_starter_monomer
from enum import Enum, unique


_PKS_TO_SMILES = {
    "WILDCARD": r"SC(C([*])C(O)=O)=O",
    "WILDCARD_STARTER": r"SC(C([*]))=O",
    "MALONYL_COA": r"SC(CC(O)=O)=O",
    "METHYLMALONYL_COA": r"SC([C@H](C)C(O)=O)=O",
    "METHOXYMALONYL_COA": r"SC(C(C(O)=O)OC)=O",
    "METHYLBUTYRYL_COA_2S": r"SC(=O)[C@@H](C)CC",
    "METHYLBUTYRYL_COA_2R": r"SC(=O)[C@H](C)CC",
    "ETHYLMALONYL_COA": r"SC(C(C(O)=O)CC)=O",
    "PROPIONYL_COA": r"SC(=O)CC",
    "ACETYL_COA": r"SC(C)=O",
    "BENZOYL_COA": r"SC(C1=CC=CC=C1)=O",
    "METHYL_BUTYRYL_COA_3": r"SC(=O)CC(C)C",
    "METHYL_BUTYRYL_COA_2": r"SC(=O)C(C)CC",
    "TRANS_CYCLOPENTANE_DICARBOXYL_COA": r"SC([C@H]1[C@@H](CCC1)C(=O)O)=O",
    "CYCLOHEXANE_CARBOXYL_COA": r"SC(=O)C1CCCCC1",
    "HYDROXY_MALONYL_COA_2": r"SC(=O)C(C(=O)O)O",
    "HYDROXY_MALONYL_COA_2R": r"SC(=O)[C@@H](C(=O)O)O",
    "HYDROXY_MALONYL_COA_2S": r"SC(=O)[C@H](C(=O)O)O",
    "CHLOROETHYL_MALONYL_COA": r"SC(C(C(=O)O)CC[Cl])=O",
    "ISOBUTYRYL_COA": r"SC(=O)C(C)C",
    "GLYCINE": "SC(CN)=O",
    "HYDROXY_PROPENOYL_COA_3_23E": r"[H]\C(O)=C/C(S)=O",
    "HYDROXY_BUTENOYL_COA_3_23E": r"C\C(O)=C/C(S)=O",
    # "DIHYDROXY_BUTANOLYL_COA_2R3": r"CC(O)[C@@H](O)C(O)S",
    # "TRIHYDROXY_PROPANOLYL_COA_233": r"OC(O)C(O)C(O)S",
    "O_METHYLACETYL_COA": r"COC(S)=O",
    "HYDROXY_PROPENOYL_COA_3_23Z": r"[H]\C(O)=C\C(S)=O",
    "OXOMALONYL_COA_2": r"OC(=O)C(=O)C(S)=O",
    "METHYL_HYDROXY_PROPENOYL_COA_2_3_23Z": r"[H]\C(O)=C(/C)C(S)=O",
    # "DIHYDROXY_BUTANOLYL_COA_23": r"CC(O)C(O)C(O)S",
    # "DIHYDROXY_BUTANOLYL_COA_2S3S": r"C[C@H](O)[C@H](O)C(O)S",
    "HEPTATRIENOYL_COA": r"SC(=O)C=CC=CC=C",
    "HYDROXYPROPIONYL_COA_2R": r"C[C@@H](O)C(S)=O",
    # "DIHYDROXY_PROPANOLYL_COA_33": r"OC(O)CC(O)S",
    "LACTYL_COA": r"C[C@@H](O)C(S)=O",
    "PHENYLACETYLCOA": r"SC(=O)CC1=CC=CC=C1",
    "METHOXYFORMYL_COA": r"COC(S)=O",
}

_TERPENE_PRECURSOR_TO_SMILES = {
    "DIMETHYLALLYL_PYROPHOSPHATE": r"CC(=CCOP(=O)(O)OP(=O)(O)O)C",
    "GERANYL_PYROPHOSPHATE": r"O=P(O)(O)OP(=O)(OC/C=C(/CC\C=C(/C)C)C)O",
    "FARNESYL_PYROPHOSPHATE": r"CC(=CCC/C(=C/CC/C(=C/COP(=O)(O)OP(=O)(O)O)/C)/C)C",
    "GERANYLGERANYL_PYROPHOSPHATE": r"O=P(O)(O)OP(=O)(O)OC/C=C(/CC\C=C(/C)CC\C=C(/C)CC\C=C(/C)C)C",
    "SQUALENE": r"CC(=CCC/C(=C/CC/C(=C/CC/C=C(/CC/C=C(/CCC=C(C)C)\C)\C)/C)/C)C",
    "PHYTOENE": r"CC(=CCC/C(=C/CC/C(=C/CC/C(=C/C=C\C=C(/C)\CC/C=C(\C)/CC/C=C(\C)/CCC=C(C)C)/C)/C)/C)C",
}


@unique
class PksStarterSubstrate(Enum):
    PROPIONYL_COA = 1
    ACETYL_COA = 2
    BENZOYL_COA = 3
    METHYL_BUTYRYL_COA_3 = 4  # 3-methylbutyryl-CoA
    METHYL_BUTYRYL_COA_2 = 5  # 2-methylbutyryl-CoA
    TRANS_CYCLOPENTANE_DICARBOXYL_COA = 6
    CYCLOHEXANE_CARBOXYL_COA = 7
    HYDROXY_MALONYL_COA_2 = 8  # 2-hydroxymalonyl-CoA
    HYDROXY_MALONYL_COA_2S = 9  # 2S-hydroxymalonyl-CoA
    HYDROXY_MALONYL_COA_2R = 10  # 2R-hydroxymalonyl-CoA
    CHLOROETHYL_MALONYL_COA = 11
    ISOBUTYRYL_COA = 12
    GLYCINE = 13
    HYDROXY_PROPENOYL_COA_3_23E = 14  # 3-hydroxy-2,3-E-propenoyl-CoA
    HYDROXY_BUTENOYL_COA_3_23E = 15  # 3-hydroxy-2,3-E-butenoyl-CoA
    O_METHYLACETYL_COA = 16  # O-methylacetyl-CoA
    HYDROXY_PROPENOYL_COA_3_23Z = 17  # 3-hydroxy-2,3-Z-propenoyl-CoA
    OXOMALONYL_COA_2 = 18  # 2-oxomalonyl-CoA
    METHYL_HYDROXY_PROPENOYL_COA_2_3_23Z = 19  # 2-methyl-3-hydroxy-2,3-Z-propenoyl-CoA
    HEPTATRIENOYL_COA = 20  # 2,4,6-heptatrienoyl-CoA
    HYDROXYPROPIONYL_COA_2R = 21  # 2R-hydroxypropionyl-CoA
    LACTYL_COA = 22
    PHENYLACETYLCOA = 23
    METHOXYFORMYL_COA = 24
    WILDCARD_STARTER = 25

    @staticmethod
    def from_string(label: str):
        for value in PksStarterSubstrate:
            if str(value) == label:
                return value
        raise ValueError(f"Unknown PKS starter substrate: {label}")

    @staticmethod
    def get_smiles(label: str):
        for value in PksStarterSubstrate:
            if str(value) == label:
                return _PKS_TO_SMILES[value]

        raise ValueError(f"Unknown PKS starter substrate: {label}")


@unique
class PksElongationSubstrate(Enum):
    WILDCARD = 1
    MALONYL_COA = 2
    METHYLMALONYL_COA = 3
    METHOXYMALONYL_COA = 4
    # TODO: something is wrong with those substrates
    # METHYLBUTYRYL_COA_2S = 5
    # METHYLBUTYRYL_COA_2R = 6
    ETHYLMALONYL_COA = 7

    @staticmethod
    def from_string(label: str):
        for value in PksElongationSubstrate:
            if str(value) == label:
                return value
        raise ValueError(f"Unknown PKS elongation substrate: {label}")

    @staticmethod
    def get_smiles(label: str):
        for value in PksElongationSubstrate:
            if str(value) == label:
                return _PKS_TO_SMILES[value]

        raise ValueError(f"Unknown PKS elongation substrate: {label}")


class Substrate:
    def __init__(self, name: str, smiles: str) -> None:
        self.name = name
        self.smiles = smiles


class NRPSSubstrate(Substrate):
    def __init__(self, name: str) -> None:
        smiles = get_smiles(name)
        super().__init__(name, smiles)


class AminoAcidSubstrate(Substrate):
    def __init__(self, name: str) -> None:
        smiles = get_smiles(name)
        super().__init__(name, smiles)


class PKSSubstrate(Substrate):
    def __init__(self, name: str) -> None:
        smiles = _PKS_TO_SMILES.get(name, None)
        if smiles is None:

            raise ValueError(f"Cannot fetch SMILES string for PKS substrate {name}.")
        super().__init__(name, smiles)
        if name == "WILDCARD":
            self.smiles = _PKS_TO_SMILES.get("WILDCARD_STARTER", None)
            self.starter_monomer = make_starter_monomer("WILDCARD_STARTER", self.smiles)
            self.elongation_monomer = make_elongation_monomer(self.name)
        elif name in [v.name for v in PksElongationSubstrate]:
            self.elongation_monomer = make_elongation_monomer(self.name)
            self.starter_monomer = None
        elif name in [v.name for v in PksStarterSubstrate]:
            self.elongation_monomer = None
            self.starter_monomer = make_starter_monomer(self.name, self.smiles)
        else:
            raise ValueError(f"PKS substrate {self.name} is not recognised by RAIChU.")


class TerpeneCyclaseSubstrate(Substrate):
    def __init__(self, name: str) -> None:
        smiles = _TERPENE_PRECURSOR_TO_SMILES.get(name, None)
        if smiles is None:
            raise ValueError(
                f"Cannot fetch SMILES string for terpene cyclase substrate {name}."
            )
        super().__init__(name, smiles)



from joblib import load
from pathlib import Path
import unittest
from raichu.reactions.nrps_tailoring_reactions import epimerize, n_methylate

TEST_FOLDER = Path(__file__).resolve().parent

NRPS_DATA = TEST_FOLDER / "test_data/nrps/marformycin/module 2.structure"


class TestNrpsTailoring(unittest.TestCase):
    def test_epimerize_glycine(self):
        pass

    def test_epimerize_tryptophan(self):
        pass
